#Vtk read/write plugin

namespace eval Vtk {
    variable vtk_version ;#version of vtk, not of this plugin
    #internal use variables
    variable GidPostElementType ;#map vtk element type number to GiD post elements
    variable GidPostElementNNode ;#map vtk element type number to GiD post element number of nodes
    variable GidPostGaussPoint1 ;#map vtk element type number to GiD post predefined internal 1 gauss point name

    variable _offsets_mesh_post ;#to associate results to the mesh ids
    variable _cell_multiple_num_equivalent_elements;#to associate results to mesh ids in case of elements that become multiple elements
    variable mesh_bounds; #to create only once each mesh for different files with different time steps
    variable percent_done
    variable stop

    variable _merge 1 ;#import options: merge to current model or delete old model+results
    variable _all_result_steps 1 ;#export options:one file by result step or only the current
    variable _all_results 1;#export options: only current result or all results
    variable _data_mode Ascii ;#export options: Ascii Binary Appended ZLib

    variable info_plugin ;#information read from .xml plugin file
    set info_plugin(path) [file dirname [info script]] ;#to know where to find the files
    if { [info exists info_plugin(Version)] } {
        set other_plugin_version $info_plugin(Version)
    }
    array set info_plugin [ReadProblemtypeXml [file join $info_plugin(path) [file tail $info_plugin(path)].xml] InfoPlugin {Name Version MinimumGiDVersion}]
    if { [info exists other_plugin_version] && $other_plugin_version != $info_plugin(Version)} {
        WarnWinText [_ "There are two versions of %s: the %s and %s inside the plugins folder. One must be deleted" $info_plugin(Name) $other_plugin_version $info_plugin(Version)]
    }
    if { [GidUtils::VersionCmp $info_plugin(MinimumGiDVersion)] < 0 } {
        WarnWinText [_ "This plugin requires GiD %s or later" $info_plugin(MinimumGiDVersion)]
    }
}

proc Vtk::SaveXmlFile { doc filename } { 
    set fail 0
    set fout [open $filename w]
    if { $fout != "" } {
        fconfigure $fout -encoding utf-8
        puts $fout {<?xml version="1.0" encoding="utf-8"?>}
        puts $fout [$doc asXML -indent 2]
        close $fout
    } else {
        set fail 1
    }
    return $fail
}

proc Vtk::XmlCreateElement { xml_parent name {attributes ""} } {
    set newNode [[$xml_parent ownerDocument] createElement $name]
    if { $attributes != "" } {
        $newNode setAttribute {*}$attributes
    }
    $xml_parent appendChild $newNode
    return $newNode
}

proc Vtk::Init { } {
    if { 0 } {
        set more_path [file join [file dirname [info script]] scripts]
        if { [lsearch $::auto_path $more_path] == -1 } {
            lappend ::auto_path $more_path
        }
    }

    variable GidPostElementType
    variable GidPostElementNNode
    variable GidPostGaussPoint1
    variable VtkElementType

    set GidPostElementType(1) Point
    #set GidPostElementType(2) Poly Point
    set GidPostElementType(3) Line
    #set GidPostElementType(4) Poly Line
    set GidPostElementType(5) Triangle
    #set GidPostElementType(6) Triangle strip
    #set GidPostElementType(7) Polygon
    set GidPostElementType(8) Quadrilateral ;#pixel
    set GidPostElementType(9) Quadrilateral
    set GidPostElementType(10) Tetrahedra
    set GidPostElementType(11) Hexahedra ;#voxel
    set GidPostElementType(12) Hexahedra
    set GidPostElementType(13) Prism
    set GidPostElementType(14) Pyramid
    set GidPostElementType(15) "" ;#Vtk pentagonal prism unsupported
    set GidPostElementType(16) "" ;#Vtk hexagonal prism unsupported
    #quadratic
    set GidPostElementType(21) Linear
    set GidPostElementType(22) Triangle
    set GidPostElementType(23) Quadrilateral
    set GidPostElementType(24) Tetrahedra
    set GidPostElementType(25) Hexahedra

    set GidPostElementNNode(1) 1
    set GidPostElementNNode(3) 2
    set GidPostElementNNode(5) 3
    set GidPostElementNNode(8) 4
    set GidPostElementNNode(9) 4
    set GidPostElementNNode(10) 4
    set GidPostElementNNode(11) 8
    set GidPostElementNNode(12) 8
    set GidPostElementNNode(13) 6
    set GidPostElementNNode(14) 5

    set GidPostElementNNode(21) 3
    set GidPostElementNNode(22) 6
    set GidPostElementNNode(23) 8
    set GidPostElementNNode(24) 10
    set GidPostElementNNode(25) 20

    set GidPostGaussPoint1(1) GP_POINT_1
    set GidPostGaussPoint1(3) GP_LINEAR_1
    #set GidPostGaussPoint1(4) Poly Line
    set GidPostGaussPoint1(5) GP_TRIANGLE_1
    #set GidPostGaussPoint1(6) Triangle strip
    #set GidPostGaussPoint1(7) Polygon
    set GidPostGaussPoint1(8) GP_QUADRILATERAL_1 ;#pixel
    set GidPostGaussPoint1(9) GP_QUADRILATERAL_1
    set GidPostGaussPoint1(10) GP_TETRAHEDRA_1
    set GidPostGaussPoint1(11) GP_HEXAHEDRA_1 ;#voxel
    set GidPostGaussPoint1(12) GP_HEXAHEDRA_1
    set GidPostGaussPoint1(13) GP_PRISM_1
    set GidPostGaussPoint1(14) GP_PYRAMID_1
    #quadratic
    set GidPostGaussPoint1(21) GP_LINEAR_1
    set GidPostGaussPoint1(22) GP_TRIANGLE_1
    set GidPostGaussPoint1(23) GP_QUADRILATERAL_1
    set GidPostGaussPoint1(24) GP_TETRAHEDRA_1
    set GidPostGaussPoint1(25) GP_HEXAHEDRA_1

    set VtkElementType(Point) 1
    set VtkElementType(Linear) 3
    set VtkElementType(Line) 3
    set VtkElementType(Triangle) 5
    set VtkElementType(Quadrilateral) 9
    set VtkElementType(Tetrahedra) 10
    set VtkElementType(Hexahedra) 12
    set VtkElementType(Prism) 13
    set VtkElementType(Pyramid) 14
    set VtkElementType(Sphere) 1 ;#export as Point element
    set VtkElementType(Circle) 1 ;#export as Point element
}

proc Vtk::GetGidPostElementType { element_code } {
    variable GidPostElementType
    return $GidPostElementType($element_code)
}

proc Vtk::GetGidPostElementNNode { element_code } {
    variable GidPostElementNNode
    return $GidPostElementNNode($element_code)
}

proc Vtk::GetGidPostGaussPoint1Type { element_code } {
    variable GidPostGaussPoint1
    return $GidPostGaussPoint1($element_code)
}

proc Vtk::GetObjArrayFromVtkDataArray { vtkDataArray } {
    return [vtk_objarray objarray_from_vtk_data_array $vtkDataArray]
}

proc Vtk::GetGetObjArrayFromVtkIdList { VtkIdList } {
    return [vtk_objarray objarray_from_vtk_data_array $VtkIdList]
}

#get list from 0 to n-1
proc Vtk::GetObjArrayIncreasingNumbers { n } {
    return [objarray new_from_to intarray 0 [expr {$n-1}]]
}

#get list increasing from start to end
proc Vtk::GetObjArrayIncreasingFromTo { start end } {
    return [objarray new_from_to intarray $start $end]
}

proc Vtk::IncrListItems { obj increment } {
    return [objarray incr $obj $increment]
}


#########################################################################################
######################################### READ POST #####################################
#########################################################################################


proc Vtk::ConvertTriangleStripToTriangles { element_nodes } {
    set n [expr {[llength $element_nodes]-2}]
    set verso 1
    set new_element_nodes {}
    for {set i 0} {$i < $n } {incr i} {
        if { $verso } {
            lappend new_element_nodes [lindex $element_nodes $i] [lindex $element_nodes $i+2] [lindex $element_nodes $i+1]
            set verso 0
        } else {
            lappend new_element_nodes [lindex $element_nodes $i] [lindex $element_nodes $i+1] [lindex $element_nodes $i+2]
            set verso 1
        }
    }
    return $new_element_nodes
}

proc Vtk::ConvertPolyLineToLines { element_nodes } {
    set new_element_nodes [lindex $element_nodes 0]
    foreach item [lrange $element_nodes 1 end-1] {
        lappend new_element_nodes $item $item
    }
    lappend new_element_nodes [lindex $element_nodes end]
    return $new_element_nodes
}

proc Vtk::ConvertPolygonToTriangles { node_ids vtkPolygon } {
    set element_nodes {}
    set outTris [vtkIdList New]
    #Triangulate will number outTris locally from 0 to n-1. Must restore after node_ids
    set cont 0
    foreach id $node_ids {
        set ids($cont) $id
        incr cont
    }
    set ok [$vtkPolygon Triangulate $outTris]
    if { $ok } {
        foreach id [Vtk::GetGetObjArrayFromVtkIdList $outTris] {
            lappend element_nodes $ids($id)
        }
    } else {
        W "polygon triangulation fail"
    }
    $outTris Delete
    return $element_nodes
}


#vtk package 5.8 fail if "<VTKFile type="UnstructuredGrid" version="1.0"..."
#                but works well in my test simply replacing 1.0 by 0.1 !!
proc Vtk::TrickToCheckVTKFileVersion { h } {
    set msg ""
    set filename [$h GetFileName]
    if { [file exists $filename] } {
        set fp [open $filename r]
        if { $fp != "" } {
            #find something like <VTKFile type="UnstructuredGrid" version="1.0" byte_order="LittleEndian" header_type="UInt64">
            gets $fp line
            if { [string range $line 0 8] == {<VTKFile } } {
                if { [regexp {version=([^ ]*)} $line dummy version] } {
                    set version [string trim $version \"]
                    set version [string trim $version ']
                    if { [string is double $version] && $version >= 1.0 } {
                        set msg "VTKFile type=UnstructuredGrid version=1.0 or greated can't be read by this vtk [package require vtk] package, try editing the file replacing version=$version by version=0.1 and reload the file"
                    }
                }
            }
            close $fp
        }
    }
    return $msg
}

#return cell_type and element_nodes changed by more than one element of other type
proc Vtk::_ConvertUnsupportedCell { i_element cell_type vtkUnstructuredGrid element_nodes } {
    if { $cell_type == 4 } {
        #polyline
        set cell_type 3 ;#line
        set element_nodes [Vtk::ConvertPolyLineToLines $element_nodes]
        set num_repetitions [expr {[llength $element_nodes]/2}]
        set element_vtk_ids [objarray new intarray $num_repetitions $i_element]
    } elseif { $cell_type == 6 } {
        #triangle strip
        set cell_type 5 ;#triangle
        set element_nodes [Vtk::ConvertTriangleStripToTriangles $element_nodes]
        set num_repetitions [expr {[llength $element_nodes]/3}]
        set element_vtk_ids [objarray new intarray $num_repetitions $i_element]
    } elseif { $cell_type == 7 } {
        #polygon
        set cell_type 5 ;#triangle
        set vtkPolygon [$vtkUnstructuredGrid GetCell $i_element]
        set element_nodes [Vtk::ConvertPolygonToTriangles $element_nodes $vtkPolygon]
        $vtkPolygon Delete
        set num_repetitions [expr {[llength $element_nodes]/3}]
        set element_vtk_ids [objarray new intarray $num_repetitions $i_element]
    } elseif { $cell_type == 8 } {
        #pixel
        set aux [objarray get $element_nodes 2]
        objarray set $element_nodes 2 [objarray get $element_nodes 3]
        objarray set $element_nodes 3 $aux
        #set element_vtk_ids $i_element
        set element_vtk_ids [objarray new intarray 1 $i_element]
    } elseif { $cell_type == 11 } {
        #voxel
        set aux [objarray get $element_nodes 2]
        objarray set $element_nodes 2 [objarray get $element_nodes 3]
        objarray set $element_nodes 3 $aux
        set aux [objarray get $element_nodes 6]
        objarray set $element_nodes 6 [objarray get $element_nodes 7]
        objarray set $element_nodes 7 $aux
        #set element_vtk_ids $i_element
        set element_vtk_ids [objarray new intarray 1 $i_element]
    } elseif { $cell_type == 13 } {
        #prism
        set aux [objarray get $element_nodes 1]
        objarray set $element_nodes 1 [objarray get $element_nodes 2]
        objarray set $element_nodes 2 $aux
        set aux [objarray get $element_nodes 4]
        objarray set $element_nodes 4 [objarray get $element_nodes 5]
        objarray set $element_nodes 5 $aux
        #set element_vtk_ids $i_element
        set element_vtk_ids [objarray new intarray 1 $i_element]
    } else {
        #set element_vtk_ids $i_element
        set element_vtk_ids [objarray new intarray 1 $i_element]
    }
    return [list $cell_type $element_nodes $element_vtk_ids]
}

proc Vtk::ReadPostUnstructuredGrid { h} {
    variable VtkElementNNode
    variable _offsets_mesh_post ;#to associate results to the mesh ids
    variable mesh_bounds; #to create only once each mesh for different files with different time steps
    variable elements_vtk_id ;#to handle result on diferent element types that must be reordered by types in GiD

    set filename [$h GetFileName]

    set num_meshes 0
    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set offset_elements [GiD_Info Mesh -post MaxNumElements]

    set vtkUnstructuredGrid [$h GetOutput]
    set vtkPoints [$vtkUnstructuredGrid GetPoints]

    if { $vtkPoints == "" } {
        $vtkUnstructuredGrid Delete
        set msg [Vtk::TrickToCheckVTKFileVersion $h]
        if { $msg == "" } {
            set msg "can't get vtkUnstructuredGrid points"
        }
        W $msg
        return $num_meshes
    }
    set gidpost_node_ids {} ;#empty for implicit increasing from the last
    set gidpost_element_ids {} ;#empty for implicit increasing from the last

    set num_nodes [$vtkPoints GetNumberOfPoints]
    set bounds [$vtkPoints GetBounds]

    set vtkDataArray [$vtkPoints GetData]
    $vtkPoints Delete
    set gidpost_nodes [vtk_objarray objarray_from_vtk_data_array $vtkDataArray]
    $vtkDataArray Delete

    set num_elements [$vtkUnstructuredGrid GetNumberOfCells]
    set vtkCellArray [$vtkUnstructuredGrid GetCells]

    set vtkIdTypeArray [$vtkCellArray GetData]
    $vtkCellArray Delete
    set data [vtk_objarray objarray_from_vtk_data_array $vtkIdTypeArray]

    $vtkIdTypeArray Delete
    set vtkUnsignedCharArray [$vtkUnstructuredGrid GetCellTypesArray]
    set cell_types [vtk_objarray objarray_from_vtk_data_array $vtkUnsignedCharArray]
    $vtkUnsignedCharArray Delete
    set different_cell_types [lsort -unique -integer $cell_types]
    #set different_cell_types [objarray clone $cell_types]
    #objarray sort -unique $different_cell_types ;#this change different_cell_types
    if { [llength $different_cell_types] == 1 } {
        #do faster this simple case
        set cell_type [lindex $cell_types 0]
        set num_nodes_element [objarray get $data 0]
        set pos 0
        for {set i_element 0 } {$i_element < $num_elements} {incr i_element} {
            set element_nodes [objarray range $data [expr {$pos+1}] [expr {$pos+$num_nodes_element}]]
            lassign [Vtk::_ConvertUnsupportedCell $i_element $cell_type $vtkUnstructuredGrid $element_nodes] cell_type_final element_nodes_final element_vtk_ids
            lappend elements($cell_type_final) {*}$element_nodes_final
            lappend elements_vtk_id($cell_type_final) {*}$element_vtk_ids
            incr pos [expr {$num_nodes_element+1}]
        }
    } else {
        for {set i_element 0 } {$i_element < $num_elements} {incr i_element} {
            set vtkCell [$vtkUnstructuredGrid GetCell $i_element]
            set vtkPointIds [$vtkCell GetPointIds]
            set element_nodes [vtk_objarray objarray_from_vtk_data_array $vtkPointIds]
            $vtkPointIds Delete
            set cell_type [$vtkCell GetCellType]
            $vtkCell Delete
            lassign [Vtk::_ConvertUnsupportedCell $i_element $cell_type $vtkUnstructuredGrid $element_nodes] cell_type_final element_nodes_final element_vtk_ids
            lappend elements($cell_type_final) {*}$element_nodes_final
            #             if { [info exists elements($cell_type_final)] } {
            #                 set elements($cell_type_final) [objarray concat $elements($cell_type_final) $element_nodes_final]
            #             } else {
            #                 set elements($cell_type_final) $element_nodes_final
            #             }
            lappend elements_vtk_id($cell_type_final) {*}$element_vtk_ids
        }
    }
    set different_cell_types [lsort -unique -integer [array names elements]]
    foreach cell_type $different_cell_types {
        if { [info exists mesh_bounds($num_nodes,$num_elements,$cell_type)] } {
            if { $mesh_bounds($num_nodes,$num_elements,$cell_type) == $bounds } {
                continue
            }
        }
        set gidpost_element_type [GetGidPostElementType $cell_type]
        if { $gidpost_element_type != "" } {
            set meshname [Vtk::GetPostMeshNameAutomatic $filename $gidpost_element_type] ;#??
            set gidpost_element_nnode [GetGidPostElementNNode $cell_type]
            Vtk::CreateGiD_MeshPost $meshname \
                $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
                $gidpost_element_ids $elements($cell_type) $offset_nodes $offset_elements
            incr num_meshes
            set _offsets_mesh_post [list $offset_nodes $offset_elements]
            #assume that is the same mesh if match num_nodes, num_elements and node's bounds
            set mesh_bounds($num_nodes,$num_elements,$cell_type) $bounds
        } else {
            W [_ "GetGidPostElementType. Unexpected or unsupported cell_type '%s'" $cell_type]
        }
    }
    $vtkUnstructuredGrid Delete
    return $num_meshes
}

proc Vtk::GetStructuredNodes3D { xs ys zs } {
    set n [expr {[llength $xs]*[llength $ys]*[llength $zs]*3}]
    set gidpost_nodes [objarray new floatarray $n 0]
    set cont 0
    foreach z $zs {
        foreach y $ys {
            foreach x $xs {
                objarray set $gidpost_nodes $cont $x
                incr cont
                objarray set $gidpost_nodes $cont $y
                incr cont
                objarray set $gidpost_nodes $cont $z
                incr cont
            }
        }
    }
    return $gidpost_nodes
}

proc Vtk::GetStructuredNodes2D { xs ys } {
    set n [expr {[llength $xs]*[llength $ys]*3}]
    set gidpost_nodes [objarray new floatarray $n 0]
    set cont 0
    foreach y $ys {
        foreach x $xs {
            objarray set $gidpost_nodes $cont $x
            incr cont
            objarray set $gidpost_nodes $cont $y
            incr cont
            objarray set $gidpost_nodes $cont 0.0
            incr cont
        }
    }
    return $gidpost_nodes
}

proc Vtk::GetStructuredNodes1D { xs } {
    set n [expr {[llength $xs]*3}]
    set gidpost_nodes [objarray new floatarray $n 0]
    set cont 0
    foreach x $xs {
        objarray set $gidpost_nodes $cont $x
        incr cont
        objarray set $gidpost_nodes $cont 0.0
        incr cont
        objarray set $gidpost_nodes $cont 0.0
        incr cont
    }
    return $gidpost_nodes
}

proc Vtk::GetStructuredConnectivities3D { n_x n_y n_z {offset_nodes 0}} {
    set n [expr {($n_x-1)*($n_y-1)*($n_z-1)*8}]
    set gidpost_elements [objarray new intarray $n 0]
    set cont 0
    set n_xy [expr {$n_x*$n_y}]
    for {set k 1} {$k < $n_z} {incr k} {
        for {set j 1} {$j < $n_y} {incr j} {
            for {set i 1} {$i < $n_x} {incr i} {
                set n0 [expr {($k-1)*$n_xy+($j-1)*$n_x+$i-1+$offset_nodes}]
                set n1 [expr {$n0+1}]
                set n2 [expr {$n1+$n_x}]
                set n3 [expr {$n2-1}]
                set n4 [expr {$n0+$n_xy}]
                set n5 [expr {$n4+1}]
                set n6 [expr {$n5+$n_x}]
                set n7 [expr {$n6-1}]
                objarray set $gidpost_elements $cont $n0
                incr cont
                objarray set $gidpost_elements $cont $n1
                incr cont
                objarray set $gidpost_elements $cont $n2
                incr cont
                objarray set $gidpost_elements $cont $n3
                incr cont
                objarray set $gidpost_elements $cont $n4
                incr cont
                objarray set $gidpost_elements $cont $n5
                incr cont
                objarray set $gidpost_elements $cont $n6
                incr cont
                objarray set $gidpost_elements $cont $n7
                incr cont
            }
        }
    }
    return $gidpost_elements
}

proc Vtk::GetStructuredConnectivities2D { n_x n_y {offset_nodes 0}} {
    set n [expr {($n_x-1)*($n_y-1)*4}]
    set gidpost_elements [objarray new intarray $n 0]
    set cont 0
    set n_xy [expr {$n_x*$n_y}]
    for {set j 1} {$j < $n_y} {incr j} {
        for {set i 1} {$i < $n_x} {incr i} {
            set n0 [expr {($j-1)*$n_x+$i-1+$offset_nodes}]
            set n1 [expr {$n0+1}]
            set n3 [expr {$n0+$n_x}]
            set n2 [expr {$n3+1}]
            objarray set $gidpost_elements $cont $n0
            incr cont
            objarray set $gidpost_elements $cont $n1
            incr cont
            objarray set $gidpost_elements $cont $n2
            incr cont
            objarray set $gidpost_elements $cont $n3
            incr cont
        }
    }
    return $gidpost_elements
}

proc Vtk::GetStructuredConnectivities1D { n_x {offset_nodes 0}} {
    set n [expr {($n_x-1)*2}]
    set gidpost_elements [objarray new intarray $n 0]
    set cont 0
    for {set i 1} {$i < $n_x} {incr i} {
        set n0 [expr {$i-1+$offset_nodes}]
        set n1 [expr {$n0+1}]
        objarray set $gidpost_elements $cont $n0
        incr cont
        objarray set $gidpost_elements $cont $n1
        incr cont
    }
    return $gidpost_elements
}

proc Vtk::ReadPostStructuredGrid { h} {
    variable _offsets_mesh_post ;#to associate results to the mesh ids
    #variable mesh_bounds; #to create only once each mesh for different files with different time steps

    set filename [$h GetFileName]
    set num_meshes 0
    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set offset_elements [GiD_Info Mesh -post MaxNumElements]

    set vtkStructuredGrid [$h GetOutput]
    set dimensions [$vtkStructuredGrid GetDimensions]
    lassign $dimensions n_x n_y n_z
    set vtkPoints [$vtkStructuredGrid GetPoints]
    $vtkStructuredGrid Delete

    set gidpost_node_ids {} ;#empty for implicit increasing from the last
    set gidpost_element_ids {} ;#empty for implicit increasing from the last

    set num_nodes [$vtkPoints GetNumberOfPoints]
    set bounds [$vtkPoints GetBounds]

    set vtkDataArray [$vtkPoints GetData]
    $vtkPoints Delete
    set gidpost_nodes [vtk_objarray objarray_from_vtk_data_array $vtkDataArray]
    $vtkDataArray Delete

    if { $n_z > 1 } {
        set gidpost_element_type Hexahedra
        set gidpost_element_nnode 8
        set gidpost_elements [Vtk::GetStructuredConnectivities3D $n_x $n_y $n_z $offset_nodes]
    } else {
        if { $n_y > 1 } {
            set gidpost_element_type Quadrilateral
            set gidpost_element_nnode 4
            set gidpost_elements [Vtk::GetStructuredConnectivities2D $n_x $n_y $offset_nodes]
        } else {
            if { $n_x > 1 } {
                set gidpost_element_type Line
                set gidpost_element_nnode 2
                set gidpost_elements [Vtk::GetStructuredConnectivities1D $n_x $offset_nodes]
            } else {
                set gidpost_element_type ""
                set gidpost_element_nnode 0
                set gidpost_elements {}
                W [_ "Grid with dimension zero"]
            }
        }
    }

    set meshname [Vtk::GetPostMeshNameAutomatic $filename $gidpost_element_type] ;#??
    GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
        $gidpost_element_ids $gidpost_elements
    incr num_meshes
    set _offsets_mesh_post [list $offset_nodes $offset_elements]
    ##assume that is the same mesh if match num_nodes, num_elements and node's bounds
    #set mesh_bounds($num_nodes,$num_elements,$cell_type) $bounds
    return $num_meshes
}


proc Vtk::ReadPostRectilinearGrid { h} {
    variable _offsets_mesh_post ;#to associate results to the mesh ids
    #variable mesh_bounds; #to create only once each mesh for different files with different time steps

    set filename [$h GetFileName]
    set num_meshes 0
    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set offset_elements [GiD_Info Mesh -post MaxNumElements]

    set vtkRectilinearGrid [$h GetOutput]
    set dimensions [$vtkRectilinearGrid GetDimensions]
    lassign $dimensions n_x n_y n_z
    set vtkFloatArray_x [$vtkRectilinearGrid GetXCoordinates]
    set vtkFloatArray_y [$vtkRectilinearGrid GetYCoordinates]
    set vtkFloatArray_z [$vtkRectilinearGrid GetZCoordinates]
    set bounds [list {*}[$vtkFloatArray_x GetValueRange] {*}[$vtkFloatArray_y GetValueRange] {*}[$vtkFloatArray_z GetValueRange]]

    $vtkRectilinearGrid Delete

    set gidpost_node_ids {} ;#empty for implicit increasing from the last
    set gidpost_element_ids {} ;#empty for implicit increasing from the last

    set num_nodes [expr {$n_x*$n_y*$n_z}]
    set xs [Vtk::GetObjArrayFromVtkDataArray $vtkFloatArray_x]
    set ys [Vtk::GetObjArrayFromVtkDataArray $vtkFloatArray_y]
    set zs [Vtk::GetObjArrayFromVtkDataArray $vtkFloatArray_z]

    $vtkFloatArray_x Delete
    $vtkFloatArray_y Delete
    $vtkFloatArray_z Delete
    set gidpost_nodes [Vtk::GetStructuredNodes3D $xs $ys $zs]

    if { $n_z > 1 } {
        set gidpost_element_type Hexahedra
        set gidpost_element_nnode 8
        set gidpost_elements [Vtk::GetStructuredConnectivities3D $n_x $n_y $n_z $offset_nodes]
    } else {
        if { $n_y > 1 } {
            set gidpost_element_type Quadrilateral
            set gidpost_element_nnode 4
            set gidpost_elements [Vtk::GetStructuredConnectivities2D $n_x $n_y $offset_nodes]
        } else {
            if { $n_x > 1 } {
                set gidpost_element_type Line
                set gidpost_element_nnode 2
                set gidpost_elements [Vtk::GetStructuredConnectivities1D $n_x $offset_nodes]
            } else {
                set gidpost_element_type ""
                set gidpost_element_nnode 0
                set gidpost_elements {}
                W [_ "Grid with dimension zero"]
            }
        }
    }

    set meshname [Vtk::GetPostMeshNameAutomatic $filename $gidpost_element_type] ;#??
    GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
        $gidpost_element_ids $gidpost_elements
    incr num_meshes
    set _offsets_mesh_post [list $offset_nodes $offset_elements]
    ##assume that is the same mesh if match num_nodes, num_elements and node's bounds
    #set mesh_bounds($num_nodes,$num_elements,$cell_type) $bounds
    return $num_meshes
}

#get a list of n equispaced real numbers from x0 to x0+(n-1)*delta
proc Vtk::GetLinearNumbers { n x0 delta } {
    set result {}
    for {set i 0} {$i < $n} {incr i} {
        lappend result [expr {$x0 + $i * $delta}]
    }
    return $result
}

proc Vtk::ReadPostStructuredPoints { h} {
    variable _offsets_mesh_post ;#to associate results to the mesh ids
    #variable mesh_bounds; #to create only once each mesh for different files with different time steps

    set filename [$h GetFileName]
    set num_meshes 0
    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set offset_elements [GiD_Info Mesh -post MaxNumElements]

    set vtkStructuredPoints [$h GetOutput]
    set dimensions [$vtkStructuredPoints GetDimensions]
    set origin [$vtkStructuredPoints GetOrigin]
    set spacing [$vtkStructuredPoints GetSpacing]
    $vtkStructuredPoints Delete
    lassign $dimensions n_x n_y n_z

    foreach axe {x y z} n $dimensions min $origin delta $spacing {
        set ${axe}s [Vtk::GetLinearNumbers $n $min $delta]
    }

    if { $n_z > 1 } {
        Vtk::CreatePostMeshGrid3D $xs $ys $zs data3d
    } else {
        if { $n_y > 1 } {
            Vtk::CreatePostMeshGrid2D $xs $ys data2d
        } else {
            if { $n_x > 1 } {
                Vtk::CreatePostMeshGrid1D $xs data1d
            } else {
                W [_ "Grid with dimension zero"]
            }
        }
    }
    incr num_meshes
    set _offsets_mesh_post [list $offset_nodes $offset_elements]
    ##assume that is the same mesh if match num_nodes, num_elements and node's bounds
    #set mesh_bounds($num_nodes,$num_elements,$cell_type) $bounds
    return $num_meshes
}

proc Vtk::ReadPostPolyData { h} {
    variable _offsets_mesh_post ;#to associate results to the mesh ids
    variable mesh_bounds; #to create only once each mesh for different files with different time steps

    variable _cell_multiple_num_equivalent_elements;#to associate results to mesh ids in case of elements that become multiple elements

    set filename [$h GetFileName]
    set num_meshes 0
    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set offset_elements [GiD_Info Mesh -post MaxNumElements]

    set vtkPolyData [$h GetOutput]
    set vtkPoints [$vtkPolyData GetPoints]
    if { $vtkPoints == "" } {
        $vtkPolyData Delete
        return 0
    }

    set gidpost_node_ids {} ;#empty for implicit increasing from the last
    set gidpost_element_ids {} ;#empty for implicit increasing from the last


    set num_nodes [$vtkPoints GetNumberOfPoints]
    set bounds [$vtkPoints GetBounds]


    set vtkDataArray [$vtkPoints GetData]
    $vtkPoints Delete
    set gidpost_nodes [vtk_objarray objarray_from_vtk_data_array $vtkDataArray]
    $vtkDataArray Delete

    set num_elements_ini 0
    set num_elements [$vtkPolyData GetNumberOfVerts]
    if { $num_elements } {
        set cell_type_final 1 ;#1-point element, constant for all
        set nnode_element_final 1

        set vtkCellArray [$vtkPolyData GetVerts]
        set vtkIdTypeArray [$vtkCellArray GetData]
        $vtkCellArray Delete
        set data [vtk_objarray objarray_from_vtk_data_array $vtkIdTypeArray]
        $vtkIdTypeArray Delete
        set pos 0
        for {set i_element 0 } {$i_element < $num_elements} {incr i_element} {
            set i_element_polydata [expr {$num_elements_ini+$i_element}]
            #set cell_type [$vtkPolyData GetCellType $i_element_polydata]
            set num_nodes_element [lindex $data $pos]
            set element_nodes [lrange $data $pos+1 $pos+$num_nodes_element]
            lappend elements($cell_type_final) {*}$element_nodes
            incr pos [expr {$num_nodes_element+1}]
        }
    }
    incr num_elements_ini $num_elements

    set num_elements [$vtkPolyData GetNumberOfLines]
    if { $num_elements } {
        set cell_type_final 3 ;#line
        set nnode_element_final 2
        set vtkCellArray [$vtkPolyData GetLines]
        set vtkIdTypeArray [$vtkCellArray GetData]
        $vtkCellArray Delete
        set data [vtk_objarray objarray_from_vtk_data_array $vtkIdTypeArray]
        $vtkIdTypeArray Delete
        set pos 0
        for {set i_element 0 } {$i_element < $num_elements} {incr i_element} {
            set i_element_polydata [expr {$num_elements_ini+$i_element}]
            set cell_type [$vtkPolyData GetCellType $i_element_polydata]
            set num_nodes_element [lindex $data $pos]
            set element_nodes [lrange $data $pos+1 $pos+$num_nodes_element]
            if { $cell_type == 4 } {
                #polyline
                set element_nodes [Vtk::ConvertPolyLineToLines $element_nodes]
                set num_equivalent_elements_local [expr {[llength $element_nodes]/2}]
                lappend _cell_multiple_num_equivalent_elements $num_equivalent_elements_local
            } else {
                set num_equivalent_elements_local 1
            }
            lappend elements($cell_type_final) {*}$element_nodes
            incr pos [expr {$num_nodes_element+1}]
        }
    }
    incr num_elements_ini $num_elements

    set num_elements [$vtkPolyData GetNumberOfStrips]
    if { $num_elements } {
        set cell_type_final 5 ;#triangle
        set nnode_element_final 3
        set vtkCellArray [$vtkPolyData GetStrips]
        set vtkIdTypeArray [$vtkCellArray GetData]
        $vtkCellArray Delete
        set data [vtk_objarray objarray_from_vtk_data_array $vtkIdTypeArray]
        $vtkIdTypeArray Delete
        set pos 0
        for {set i_element 0 } {$i_element < $num_elements} {incr i_element} {
            set i_element_polydata [expr {$num_elements_ini+$i_element}]
            set cell_type [$vtkPolyData GetCellType $i_element_polydata]
            set num_nodes_element [lindex $data $pos]
            set element_nodes [lrange $data $pos+1 $pos+$num_nodes_element]
            if { $cell_type == 6 } {
                #triangle strip
                set element_nodes [Vtk::ConvertTriangleStripToTriangles $element_nodes]
                set num_equivalent_elements_local [expr {[llength $element_nodes]/3}]
                lappend _cell_multiple_num_equivalent_elements $num_equivalent_elements_local
            } else {
                set num_equivalent_elements_local 1
            }
            lappend elements($cell_type_final) {*}$element_nodes
            incr pos [expr {$num_nodes_element+1}]
        }
    }
    incr num_elements_ini $num_elements

    set num_elements [$vtkPolyData GetNumberOfPolys]
    if { $num_elements } {
        #cell type could be variable:e.g. a mix of triangles and quadrilaterals
        set vtkCellArray [$vtkPolyData GetPolys]
        set vtkIdTypeArray [$vtkCellArray GetData]
        $vtkCellArray Delete
        set data [vtk_objarray objarray_from_vtk_data_array $vtkIdTypeArray]
        $vtkIdTypeArray Delete
        set pos 0
        for {set i_element 0 } {$i_element < $num_elements} {incr i_element} {
            set num_nodes_element [lindex $data $pos]
            set element_nodes [lrange $data $pos+1 $pos+$num_nodes_element]
            incr pos
            incr pos $num_nodes_element
            set i_element_polydata [expr {$num_elements_ini+$i_element}]
            set cell_type [$vtkPolyData GetCellType $i_element_polydata]
            if { $cell_type == 7 } {
                set vtkPolygon [$vtkPolyData GetCell $i_element_polydata]
                set element_nodes [Vtk::ConvertPolygonToTriangles $element_nodes $vtkPolygon]
                $vtkPolygon Delete
                set cell_type_final 5 ;#split polygon into triangles
                set nnode_element_final 3
                set num_equivalent_elements_local [expr {[llength $element_nodes]/3}]
                lappend _cell_multiple_num_equivalent_elements $num_equivalent_elements_local
            } else {
                set cell_type_final $cell_type
                set nnode_element_final $num_nodes_element
                set num_equivalent_elements_local 1
            }
            if { $element_nodes == "" } {
                W "ReadPostPolyData. problems triangulating polygon ${i_element}:"
                W [string range $data 0 1000]
                foreach v [ list num_nodes_element element_nodes cell_type] {
                    W "$v = [ subst $$v]"
                }
                break
            }
            lappend elements($cell_type_final) {*}$element_nodes
        }
    }
    incr num_elements_ini $num_elements

    set different_cell_types [lsort -unique -integer [array names elements]]
    foreach cell_type $different_cell_types {
        set num_elements [llength $elements($cell_type)]
        if { [info exists mesh_bounds($num_nodes,$num_elements,$cell_type)] } {
            if { $mesh_bounds($num_nodes,$num_elements,$cell_type) == $bounds } {
                continue
            }
        }
        set gidpost_element_type [GetGidPostElementType $cell_type]
        if { $gidpost_element_type != "" } {
            set meshname [Vtk::GetPostMeshNameAutomatic $filename $gidpost_element_type] ;#??
            set gidpost_element_nnode [GetGidPostElementNNode $cell_type]
            Vtk::CreateGiD_MeshPost $meshname \
                $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
                $gidpost_element_ids $elements($cell_type) $offset_nodes $offset_elements
            incr num_meshes
            set _offsets_mesh_post [list $offset_nodes $offset_elements]
            #assume that is the same mesh if match num_nodes, num_elements and node's bounds
            set mesh_bounds($num_nodes,$num_elements,$cell_type) $bounds
        } else {
            W [_ "GetGidPostElementType. Unexpected or unsupported cell_type '%s'" $cell_type]
        }
    }

    $vtkPolyData Delete
    return $num_meshes
}

proc Vtk::CreateGiD_MeshPost { meshname gidpost_element_type gidpost_element_nnode zero_based_array \
    gidpost_node_ids gidpost_nodes gidpost_element_ids gidpost_elements {offset_nodes 0} {offset_elements 0}} {

    if { $offset_nodes } {
        # check if gipost_node_ids exists, if not --> create it
        if { [ GidUtils::IsObjarray $gidpost_node_ids] == 0} {
            set num_nodes [ expr [ objarray length $gidpost_nodes] / 3]
            set gidpost_node_ids [ objarray new_from_to intarray 0 [ expr $num_nodes - 1]]
        }
        set gidpost_node_ids [Vtk::IncrListItems $gidpost_node_ids $offset_nodes]
        set gidpost_elements [Vtk::IncrListItems $gidpost_elements $offset_nodes]
    }
    if { $offset_elements } {
        # check if gipost_node_ids exists, if not --> create it
        if {  [ GidUtils::IsObjarray $gidpost_element_ids] == 0} {
            set num_elements [ expr  [ llength $gidpost_elements] / $gidpost_element_nnode]
            set gidpost_element_ids [ objarray new_from_to intarray 0 [ expr $num_elements - 1]]
        }
        set gidpost_element_ids [Vtk::IncrListItems $gidpost_element_ids $offset_elements]
    }

    if { $zero_based_array == "-zero_based_array" } {
        GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
            $gidpost_element_ids $gidpost_elements
    } else {
        GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode $gidpost_node_ids $gidpost_nodes \
            $gidpost_element_ids $gidpost_elements
    }
}

proc Vtk::ReadPostMeshes { filename } {
    if { ![file exists $filename] } {
        W [_ "File %s not found" $filename]
        return 0
    }
    set is_vtk_dataset_reader 0
    set extension [string tolower [file extension $filename]]
    if { $extension == ".vtu" } {
        set h [vtkXMLUnstructuredGridReader New]
    } elseif { $extension == ".vtp" } {
        set h [vtkXMLPolyDataReader New]
    } elseif { $extension == ".vts" } {
        set h [vtkXMLStructuredGridReader New]
    } elseif { $extension == ".vtr" } {
        set h [vtkXMLRectilinearGridReader New]
    } elseif { $extension == ".vti" } {
        set h [vtkXMLImageDataReader New]
    } elseif { $extension == ".vto" } {
        set h [vtkXMLHyperOctreeReader New]
    } else {
        #.vtk
        set h [vtkDataSetReader New]
        set is_vtk_dataset_reader 1
    }

    $h SetFileName $filename
    $h GlobalWarningDisplayOff ;#to avoid error message with file 'financial.vtk' vtkStreamingDemandDrivenPipeline : Algorithm vtkDataSetReader returned failure for request: vtkInformation
    if { $is_vtk_dataset_reader } {
        foreach item {Scalars Vectors Normals Tensors TCoords ColorScalars Fields} {
            $h SetReadAll$item 1
        }
    }
    $h Update

    set vtkOutput [$h GetOutput]
    if { $vtkOutput != "" } {
        #some vtk files provide vtkOutput ""
        $vtkOutput GlobalReleaseDataFlagOn ;#for efficiency, to release intermediate results between filters in a pipeline
    }

    if { [$h GetErrorCode] == 20002 } {
        W [_ "File can't be opened"]
    }

    set num_meshes 0
    if { [string tolower [file extension $filename]] == ".vtu" } {
        incr num_meshes [Vtk::ReadPostUnstructuredGrid $h]
    } elseif { [string tolower [file extension $filename]] == ".vtp" } {
        incr num_meshes [Vtk::ReadPostPolyData $h]
    } elseif { [string tolower [file extension $filename]] == ".vts" } {
        incr num_meshes [Vtk::ReadPostStructuredGrid $h]
    } elseif { [string tolower [file extension $filename]] == ".vtr" } {
        incr num_meshes [Vtk::ReadPostRectilinearGrid $h]
    } elseif { [string tolower [file extension $filename]] == ".vti" } {
        incr num_meshes [Vtk::ReadPostStructuredPoints $h]
    } elseif { [string tolower [file extension $filename ]] == ".vto" } {
        #not implemented yet
    } else {
        if { [$h IsFileUnstructuredGrid] } {
            incr num_meshes [Vtk::ReadPostUnstructuredGrid $h]
        } elseif { [$h IsFilePolyData] } {
            incr num_meshes [Vtk::ReadPostPolyData $h]
        } elseif { [$h IsFileStructuredGrid]  } {
            incr num_meshes [Vtk::ReadPostStructuredGrid $h]
        } elseif { [$h IsFileRectilinearGrid] } {
            incr num_meshes [Vtk::ReadPostRectilinearGrid $h]
        } elseif { [$h IsFileStructuredPoints] } {
            incr num_meshes [Vtk::ReadPostStructuredPoints $h]
        } else {
            #W [_ "File without mesh or mesh type unknown"]
        }
    }

    $h Delete
    return $num_meshes
}

#to get an unused name, based on "prefix label"
proc Vtk::GetPostMeshNameAutomatic { filename label } {
    set prefix [ file rootname [ file tail $filename]]
    set new_name "${prefix} ${label}"
    set names [concat [GiD_Info postprocess get all_surfacesets] [GiD_Info postprocess get all_volumesets]]
    if { [lsearch $names $new_name] == -1 } {
        set name $new_name
    } else {
        set i 1
        while { 1 } {
            set name ${new_name}_$i
            if { [lsearch $names $name] == -1 } {
                break
            }
            incr i
        }
    }
    return $name
}

proc Vtk::ResultNameExists { label } {
    set exists 0
    set analysis [GiD_Info postprocess get cur_analysis]
    if { $analysis != "" } {
        set time_step [GiD_Info postprocess get cur_step]
        if { $time_step != "" } {
            set names [GiD_Info postprocess get results_list CONTOUR_FILL $analysis $time_step]
            if { [lsearch $names $label] != -1 } {
                set exists 1
            }
        }
    }
    return $exists
}

proc Vtk::GetPostResultNameAutomatic { label } {
    set analysis [GiD_Info postprocess get cur_analysis]
    set name $label
    if { $analysis != "" } {
        set time_step [GiD_Info postprocess get cur_step]
        set names [GiD_Info postprocess get results_list CONTOUR_FILL $analysis $time_step]
        if { [lsearch $names $label] == -1 } {
            set name $label
        } else {
            set i 1
            while { 1 } {
                set name ${label}_$i
                if { [lsearch $names $name] == -1 } {
                    break
                }
                incr i
            }
        }
    }
    return $name
}

proc Vtk::CreatePostMeshGrid2D { xs ys label } {
    set n_x [llength $xs]
    set n_y [llength $ys]
    if { $n_x<2 || $n_y<2 } {
        W [_ "CreatePostMeshGrid2D: xs and ys must have at least two items"]
        return 0
    }

    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set meshname $label
    set gidpost_element_type Quadrilateral
    set gidpost_element_nnode 4
    set gidpost_element_ids {} ;#empty for implicit increasing from the last

    if { [GidUtils::VersionCmp 10.1.9d] >= 0} {
        set gidpost_node_ids {} ;#empty for implicit increasing from the last
    } else {
        #error in previous versions, can't use without node ids because every mesh start by 0 !!
        set gidpost_node_ids [Vtk::GetObjArrayIncreasingFromTo [expr {$offset_nodes}] [expr {[llength $xs]*[llength $ys]-1+$offset_nodes}]]
    }

    set gidpost_nodes [Vtk::GetStructuredNodes2D $xs $ys]
    set gidpost_elements [Vtk::GetStructuredConnectivities2D $n_x $n_y $offset_nodes]

    GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
        $gidpost_element_ids $gidpost_elements
    return 1
}

proc Vtk::CreatePostMeshGrid1D { xs label } {
    set n_x [llength $xs]
    if { $n_x<2 } {
        W [_ "CreatePostMeshGrid1D: xs must have at least two items"]
        return 0
    }

    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set meshname $label
    set gidpost_element_type Line
    set gidpost_element_nnode 2
    set gidpost_element_ids {} ;#empty for implicit increasing from the last

    if { [GidUtils::VersionCmp 10.1.9d] >= 0} {
        set gidpost_node_ids {} ;#empty for implicit increasing from the last
    } else {
        #error in previous versions, can't use without node ids because every mesh start by 0 !!
        set gidpost_node_ids [Vtk::GetObjArrayIncreasingFromTo [expr {$offset_nodes}] [expr {[llength $xs]-1+$offset_nodes}]]
    }

    set gidpost_nodes [Vtk::GetStructuredNodes1D $xs]
    set gidpost_elements [Vtk::GetStructuredConnectivities1D $n_x $offset_nodes]

    GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
        $gidpost_element_ids $gidpost_elements
    return 1
}


#gidpost_type must be Scalar, ComplexScalar, Vector, ComplexVector
proc Vtk::CreatePostGraph2D { xs ys label_x label_y unit_x unit_y label unit gidpost_values gidpost_type } {
    variable _grids
    if { $gidpost_type != "Scalar" && $gidpost_type != "ComplexScalar" && $gidpost_type != "Vector" && $gidpost_type != "ComplexVector"} {
        W [_ "CreatePostGraph2D type '$gidpost_type' not implemented yet"]
        return 0
    }

    set checksum [encrypter MD5 encode [list $xs $ys]]
    if { [info exists _grids($checksum)] } {
        #not create again the mesh was created with the same xs yz
        set offset_nodes $_grids($checksum)
    } else {
        set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
        set num_meshes [Vtk::CreatePostMeshGrid2D $xs $ys [Vtk::GetPostMeshNameAutomatic "Graph2D" $label]]
        if { $num_meshes != 1 } {
            return 0
        }
        set _grids($checksum) $offset_nodes
    }

    set gidpost_ids [Vtk::GetObjArrayIncreasingFromTo [expr {1+$offset_nodes}] [expr {[llength $xs]*[llength $ys]+$offset_nodes}]]

    set gidpost_resultname $label
    set gidpost_analysisname none
    set gidpost_step 1
    set gidpost_over OnNodes
    set gidpost_result [list Result $gidpost_resultname $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
    #set gidpost_component_names [list $label_x $label_y]
    set unit_name $unit


    if { $gidpost_type == "ComplexScalar" && [GidUtils::VersionCmp "10.1.9d"] < 0 } {
        #trick to work also with versions prior to 10.1.9d that have a bug with -array for ComplexScalar
        set gidpost_some_values ""
        foreach id $gidpost_ids re [lindex $gidpost_values 0] im [lindex $gidpost_values 1] {
            lappend gidpost_some_values [list $id $re $im]
        }
        GiD_Result create $gidpost_result [list Unit $unit_name] {*}$gidpost_some_values
        return 1
    }

    set gidpost_some_values [list $gidpost_ids $gidpost_values]
    #GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names] [list Unit $unit_name] $gidpost_some_values
    GiD_Result create -array $gidpost_result [list Unit $unit_name] $gidpost_some_values
    return 1
}


proc Vtk::CreatePostMeshGrid3D { xs ys zs label } {
    if { [GidUtils::VersionCmp "10.1.9d"] < 0 } {
        W [_ "This version %s of GiD is not able to create a hexahedra grid, update GiD" [GiD_Info GiDVersion] ]
        return 0
    }
    set n_x [llength $xs]
    set n_y [llength $ys]
    set n_z [llength $zs]

    if { $n_x<2 || $n_y<2 || $n_z<2 } {
        W [_ "CreatePostMeshGrid3D: xs, ys and zs must have at least two items"]
        return 0
    }

    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set meshname $label
    set gidpost_element_type Hexahedra
    set gidpost_element_nnode 8
    set gidpost_node_ids {} ;#empty for implicit increasing from the last
    set gidpost_element_ids {} ;#empty for implicit increasing from the last

    set gidpost_nodes [Vtk::GetStructuredNodes3D $xs $ys $zs]
    set gidpost_elements [Vtk::GetStructuredConnectivities3D $n_x $n_y $n_z $offset_nodes]
    GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode -zero_based_array $gidpost_node_ids $gidpost_nodes \
        $gidpost_element_ids $gidpost_elements
    return 1
}

proc Vtk::CreatePostGraph3D { xs ys zs label_x label_y label_z unit_x unit_y unit_z label unit gidpost_values gidpost_type } {
    variable _grids
    if { $gidpost_type != "Scalar" && $gidpost_type != "ComplexScalar" && $gidpost_type != "Vector" && $gidpost_type != "ComplexVector"} {
        W [_ "CreatePostGraph3D type '%s' not implemented yet" $gidpost_type]
        return 0
    }

    set checksum [encrypter MD5 encode [list $xs $ys $zs]]
    if { [info exists _grids($checksum)] } {
        #not create again the mesh was created with the same xs yz zs
        set offset_nodes $_grids($checksum)
    } else {
        set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
        set num_meshes [Vtk::CreatePostMeshGrid3D $xs $ys $zs [Vtk::GetPostMeshNameAutomatic "Graph3D" $label]]
        if { $num_meshes != 1 } {
            return 0
        }
        set _grids($checksum) $offset_nodes
    }

    set gidpost_ids [Vtk::GetObjArrayIncreasingFromTo [expr {1+$offset_nodes}] [expr {[llength $xs]*[llength $ys]*[llength $zs]+$offset_nodes}]]

    set gidpost_resultname $label
    set gidpost_analysisname none
    set gidpost_step 1
    set gidpost_over OnNodes
    set gidpost_result [list Result $gidpost_resultname $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
    #set gidpost_component_names [list $label_x $label_y $label_z]
    set unit_name $unit

    if { $gidpost_type == "ComplexScalar" && [GidUtils::VersionCmp "10.1.9d"] < 0 } {
        #trick to work also with versions prior to 10.1.9d that have a bug with -array for ComplexScalar
        set gidpost_some_values ""
        foreach id $gidpost_ids re [lindex $gidpost_values 0] im [lindex $gidpost_values 1] {
            lappend gidpost_some_values [list $id $re $im]
        }
        GiD_Result create $gidpost_result [list Unit $unit_name] {*}$gidpost_some_values
        return 1
    }

    set gidpost_some_values [list $gidpost_ids $gidpost_values]
    #GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names] [list Unit $unit_name] $gidpost_some_values
    GiD_Result create -array $gidpost_result [list Unit $unit_name] $gidpost_some_values
    return 1
}

proc Vtk::ReadPostResultFieldDataAsGraphs { h } {
    variable vtk_version
    set num_results 0
    set num_graphs 0
    if { [$h GetClassName] != "vtkDataSetReader" } {
        return [list $num_results $num_graphs]
    }
    set unit_x ""
    set unit_y ""
    set label_x [_ "Step"]
    set label_y [_ "Value"]

    set num_fields [$h GetNumberOfFieldDataInFile]


    #$h SetReadAllFields 1
    #$h ReadAllFieldsOn
    #$h Update
    #set vtkFieldData [$h ReadFieldData]
    for {set i_field 0} {$i_field < $num_fields} {incr i_field} {
        set field_name [$h GetFieldDataNameInFile $i_field]
        $h SetFieldDataName $field_name
        $h Update
        if { $vtk_version == 5.8 } {
            set vtkFieldData [$h ReadFieldData]
        } else {
            #set FIELD_DATA 2
            set vtkFieldData [$h ReadFieldData 2]
        }
        if { $vtkFieldData == "" } {
            W "ReadPostResultFieldDataAsGraphs. ReadFieldData NULL reading field $i_field"
            break
        }
        set num_arrays_dataset [$vtkFieldData GetNumberOfArrays]
        for { set i_result 0 } { $i_result < $num_arrays_dataset } { incr i_result } {
            if { $num_fields == 1 } {
                set gidpost_resultname [$vtkFieldData GetArrayName $i_result]
            } else {
                set gidpost_resultname ${field_name}:[$vtkFieldData GetArrayName $i_result]
            }
            set vtkDataArray [$vtkFieldData GetArray $i_result]
            set num_components [$vtkDataArray GetNumberOfComponents]
            set num_tuples [$vtkDataArray GetNumberOfTuples]
            set xs [Vtk::GetObjArrayIncreasingNumbers $num_tuples]
            set data [Vtk::GetObjArrayFromVtkDataArray $vtkDataArray]
            for { set i_component 0 } { $i_component < $num_components } { incr i_component } {
                if { $num_components == 1 } {
                    set graph_name $gidpost_resultname
                    set ys $data
                } else {
                    if { $num_components == 3 } {
                        set letters {x y z}
                        set graph_name $gidpost_resultname-[lindex $letters $i_component]
                    } else {
                        set graph_name ${gidpost_resultname}-${i_component}
                    }
                    set ys [lrange $data [expr {$num_tuples*$i_component}] [expr {$num_tuples*($i_component+1)-1}]]
                }
                GiD_Graph create $graph_name $label_x $label_y $xs $ys $unit_x $unit_y
                incr num_graphs
            }
            $vtkDataArray Delete
        }
        $vtkFieldData Delete
    }
    #$vtkFieldData Delete
    return [list $num_results $num_graphs]
}

#some vtk cell types are splitted in other types
proc Vtk::GetCellTypeFinal { cell_type } {
    if { $cell_type == 4 } {
        #polyline. converted to lines
        set cell_type_final 3
    } elseif { $cell_type == 6 } {
        #triangle strip. converted to triangles
        set cell_type_final 5
    } elseif { $cell_type == 7 } {
        #polygon. converted to triangles
        set cell_type_final 5
    } else {
        set cell_type_final $cell_type
    }
    return $cell_type_final
}

proc Vtk::ReadPostResultArraySet { h analysis time_step } {
    variable _merge
    variable _offsets_mesh_post
    variable _cell_multiple_num_equivalent_elements;#to associate results to mesh ids in case of elements that become multiple elements
    variable elements_vtk_id ;#to handle result on diferent element types that must be reordered by types in GiD

    if { ![info exists _offsets_mesh_post] } {
        #mesh not exists
        lassign [Vtk::ReadPostResultFieldDataAsGraphs $h] num_results num_graphs
        return [list $num_results $num_graphs]
    }

    lassign $_offsets_mesh_post offset_nodes offset_elements

    set num_graphs 0
    set num_results 0

    set filename [$h GetFileName]
    if { $time_step != "" } {
        set gidpost_step $time_step
    } else {
        lassign [Vtk::GetTimeStepFromName $filename] base_name gidpost_step
        #set gidpost_analysisname $base_name
    }
    if { ![string is double -strict $gidpost_step] } {
        #W "Vtk::ReadPostResultArraySet. gidpost_step=$gidpost_step is not a valid double, set to 0"
        set gidpost_step 0
    }
    set gidpost_analysisname $analysis
    if { $gidpost_analysisname == "" } {
        set gidpost_analysisname "x"
    }

    set vtkDataSet [$h GetOutput]
    set class_name [$vtkDataSet GetClassName]
    foreach type {node element} {
        if { $type == "node" } {
            set vtkPointData [$vtkDataSet GetPointData]
            set vtkDataSetAttributes $vtkPointData
            set gidpost_over OnNodes
            if { $class_name == "vtkUnstructuredGrid" } {
                set num_nodes [$vtkDataSet GetNumberOfPoints]
            } elseif { $class_name == "vtkStructuredGrid" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_nodes [expr {$n_x*$n_y*$n_z}]
            } elseif { $class_name == "vtkRectilinearGrid" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_nodes [expr {$n_x*$n_y*$n_z}]
            } elseif { $class_name == "vtkStructuredPoints" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_nodes [expr {$n_x*$n_y*$n_z}]
            } elseif { $class_name == "vtkPolyData" } {
                set num_nodes [$vtkDataSet GetNumberOfPoints]
            } elseif { $class_name == "vtkImageData" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_nodes [expr {$n_x*$n_y*$n_z}]
            } else {
                W [_ "ReadPostResultArraySet unexpected vtkDataSet type %s" [$vtkDataSet GetClassName]]
                continue
            }
            #node numbers start by 1 in GiD
            set gidpost_ids [Vtk::GetObjArrayIncreasingFromTo [expr $offset_nodes+1] [expr $offset_nodes+$num_nodes]]
        } else {
            #type == element
            set vtkCellData [$vtkDataSet GetCellData]
            set vtkDataSetAttributes $vtkCellData
            set gidpost_over OnGaussPoints
            if { $class_name == "vtkUnstructuredGrid" } {
                set num_elements [$vtkDataSet GetNumberOfCells]
            } elseif { $class_name == "vtkStructuredGrid" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_elements [expr {($n_x-1)*($n_y-1)*($n_z-1)}]
                if { $n_z == 1 } {
                    if { $n_y == 1 } {
                        set cell_type 3 ;#line
                    } else {
                        set cell_type 9 ;#quadrilateral
                    }
                } else {
                    set cell_type 12 ;#hexahedra
                }
            } elseif { $class_name == "vtkRectilinearGrid" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_elements [expr {($n_x-1)*($n_y-1)*($n_z-1)}]
                if { $n_z > 1 } {
                    set cell_type 12 ;#hexahedra
                } else {
                    if { $n_y > 1 } {
                        set cell_type 9 ;#quadrilateral
                    } else {
                        if { $n_x > 1 } {
                            set cell_type 3 ;#line
                        } else {
                            set cell_type "" ;#invalid
                        }
                    }
                }
            } elseif { $class_name == "vtkStructuredPoints" } {
                #continue ;#do not exists elements
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_elements [expr {($n_x-1)*($n_y-1)*($n_z-1)}]
                if { $n_z == 1 } {
                    if { $n_y == 1 } {
                        set cell_type 3 ;#line
                    } else {
                        set cell_type 9 ;#quadrilateral
                    }
                } else {
                    set cell_type 12 ;#hexahedra
                }
            } elseif { $class_name == "vtkPolyData" } {
                set num_elements [$vtkDataSet GetNumberOfCells]
                set vtkCell [$vtkDataSet GetCell 0];#get first element, assumed equal for all!!
                set cell_type [Vtk::GetCellTypeFinal [$vtkCell GetCellType]]
                $vtkCell Delete
            } elseif { $class_name == "vtkImageData" } {
                set dimensions [$vtkDataSet GetDimensions]
                lassign $dimensions n_x n_y n_z
                set num_elements [expr {($n_x-1)*($n_y-1)*($n_z-1)}]
                if { $n_z == 1 } {
                    if { $n_y == 1 } {
                        set cell_type 3 ;#line
                    } else {
                        set cell_type 9 ;#quadrilateral
                    }
                } else {
                    set cell_type 12 ;#hexahedra
                }
            } else {
                W [_ "ReadPostResultArraySet unexpected vtkDataSet type %s" [$vtkDataSet GetClassName]]
                continue
            }

            if { $class_name != "vtkUnstructuredGrid" } {
                #element numbers start by 1 in GiD
                set elements_vtk_id($cell_type) [Vtk::GetObjArrayIncreasingFromTo [expr $offset_elements+1] [expr $offset_elements+$num_elements]]
            }
            set different_cell_types [lsort -unique -integer [array names elements_vtk_id]]
        }
        set num_arrays_dataset [$vtkDataSetAttributes GetNumberOfArrays]
        for { set i_result 0 } { $i_result < $num_arrays_dataset } { incr i_result } {
            set gidpost_resultname [$vtkDataSetAttributes GetArrayName $i_result]
            #vtk result names could be repeated for nodes and cells
            #set gidpost_resultname [Vtk::GetPostResultNameAutomatic $gidpost_resultname]
            if { [Vtk::ResultNameExists $gidpost_resultname] } {
                if { $type == "element" } {
                    set gidpost_resultname cell_${gidpost_resultname}
                } else {
                    set gidpost_resultname nodal_${gidpost_resultname}
                }
            }
            set vtkDataArray [$vtkDataSetAttributes GetArray $i_result]
            set num_components [$vtkDataArray GetNumberOfComponents]
            set num_tuples [$vtkDataArray GetNumberOfTuples] ;#must be == num nodos/elementos??
            #W "ReadPostResultArraySet $gidpost_resultname type=[$vtkDataArray GetDataType]"
            #package vtk 5.8.0 Win x64 fail in a forum case (testvtu) because return that a type is float==10 when really is double==11
            #version Win x32 works well
            #package vtk 6.1, 7.0, 7.1 could not be used because fail loading some of its dlls in Win x32 and x64
            if { $num_components == 1 } {
                set gidpost_type Scalar
                set gidpost_component_names "x"
            } elseif { $num_components == 2 } {
                #convert to vector with z=0
                set gidpost_type Vector
                if { [$vtkDataArray HasAComponentName] } {
                    set c0 [$vtkDataArray GetComponentName 0]
                    set c1 [$vtkDataArray GetComponentName 1]
                    set gidpost_component_names [list $c0 $c1 z]
                } else {
                    set gidpost_component_names "x y z"
                }
            } elseif { $num_components == 3 } {
                set gidpost_type Vector
                if { [$vtkDataArray HasAComponentName] } {
                    set c0 [$vtkDataArray GetComponentName 0]
                    set c1 [$vtkDataArray GetComponentName 1]
                    set c2 [$vtkDataArray GetComponentName 2]
                    set gidpost_component_names [list $c0 $c1 $c2]
                } else {
                    set gidpost_component_names "x y z"
                }
            } elseif { $num_components == 9 } {
                set gidpost_type Matrix
                if { [$vtkDataArray HasAComponentName] } {
                    set c0 [$vtkDataArray GetComponentName 0]
                    set c1 [$vtkDataArray GetComponentName 1]
                    set c2 [$vtkDataArray GetComponentName 2]
                    set c3 [$vtkDataArray GetComponentName 3]
                    set c4 [$vtkDataArray GetComponentName 4]
                    set c5 [$vtkDataArray GetComponentName 5]
                    set gidpost_component_names [list $c0 $c1 $c2 $c3 $c4 $c5]
                } else {
                    set gidpost_component_names "Sxx Syy Szz Sxy Syz Sxz"
                }
            } else {
                W [_ "ReadPostResultArraySet. Unexpected num_components '%s'" $num_components]
                set gidpost_type ""
            }
            if { $gidpost_type != "" } {
                if { $num_components == 9 } {
                    set num_components_used 6
                    set components_used {0 4 8 1 5 2}
                } else {
                    set num_components_used $num_components
                    set components_used [Vtk::GetObjArrayIncreasingNumbers $num_components]
                }
                set v [vtk_objarray objarray_from_vtk_data_array $vtkDataArray]
                $vtkDataArray Delete
                if { $num_components == 1 } {
                    #do faster this simple case
                    set gidpost_values [list $v]
                } else {
                    #reorder v to separate each component in consecutive blocks
                    objarray deinterleave $v $num_components
                    set gidpost_values [list]
                    for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                        set pos0 [expr {$i_component*$num_tuples}]
                        set pos1 [expr {($i_component+1)*$num_tuples-1}]
                        set data [objarray range $v $pos0 $pos1]
                        lappend gidpost_values $data
                    }
                }
                if { $num_components == 2 } {
                    #append a third vector component of zeros
                    lappend gidpost_values [objarray new [objarray type $v] $num_tuples 0]
                }

                if { $type == "element" && [llength $different_cell_types] == 1 } {
                    #I think that in case of multiple different_cell_types this was already taken into account
                    set converted_cells 0
                    foreach cell_type_to_convert {4 6 7} {
                        if { [lsearch $different_cell_types $cell_type_to_convert] != -1 } {
                            set converted_cells 1
                        }
                    }
                    if { $converted_cells } {
                        #in case of one element that become multiple must replace the element id and its result by the true id range and repeat the value
                        set num_elements_final $num_elements
                        set cont 0
                        for {set i_element 0} {$i_element < $num_elements} {incr i_element} {
                            set vtkCell [$vtkDataSet GetCell $i_element]
                            set cell_type [$vtkCell GetCellType]
                            $vtkCell Delete
                            if { $cell_type == 4 || $cell_type == 6 || $cell_type == 7 } {
                                set num_equivalent_elements [lindex $_cell_multiple_num_equivalent_elements $cont]
                                incr cont ;#only take into account this special cells
                                for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                                    set value [lindex [lindex $gidpost_values $i_component] $i_element]
                                    lappend new_gidpost_values_component($i_component) {*}[lrepeat $num_equivalent_elements $value]
                                }
                                incr num_elements_final [expr {$num_equivalent_elements-1}]
                            } else {
                                for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                                    set value [lindex [lindex $gidpost_values $i_component] $i_element]
                                    lappend new_gidpost_values_component($i_component) $value
                                }
                            }
                        }

                        set new_gidpost_ids [Vtk::GetObjArrayIncreasingFromTo [expr $offset_elements+1] [expr $offset_elements+$num_elements_final]]
                        set gidpost_ids $new_gidpost_ids
                        set gidpost_values [list]
                        for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                            lappend gidpost_values $new_gidpost_values_component($i_component)
                        }
                    }
                }

                if { $gidpost_over == "OnGaussPoints" } {
                    if { [llength $different_cell_types] == 1 } {
                        #do faster this simple case, without need to reorder elements and results to have consecutive for GiD the same kind of elements
                        #element numbers start by 1 in GiD
                        set gidpost_ids [GetObjArrayIncreasingFromTo [expr $offset_elements+1] [expr $offset_elements+$num_elements]]
                        set gidpost_gauss_point_name GP_ELEMENT_1
                        set gidpost_result [list Result $gidpost_resultname $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
                        lappend gidpost_result $gidpost_gauss_point_name
                        set gidpost_some_values [list $gidpost_ids $gidpost_values]
                        GiD_Result create -array $gidpost_result $gidpost_some_values
                        incr num_results
                    } else {
                        set gidpost_ids_reorederd [objarray new intarray 0]
                        set gidpost_values_reorederd [list]
                        for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                            lappend gidpost_values_reorederd [objarray new floatarray 0]
                        }
                        set element_gid_id $offset_elements
                        foreach cell_type_final $different_cell_types {
                            set num_elements_of_type [llength $elements_vtk_id($cell_type_final)]
                            set gidpost_ids_cell_type [Vtk::GetObjArrayIncreasingFromTo [expr $element_gid_id+1] [expr $element_gid_id+$num_elements_of_type]]
                            incr element_gid_id $num_elements_of_type
                            set gidpost_values_cell_type [list]
                            for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                                set component_values [lindex $gidpost_values $i_component]
                                set component_values_cell_type [objarray new floatarray $num_elements_of_type 0]
                                set index 0
                                foreach element_vtk_id $elements_vtk_id($cell_type_final) {
                                    objarray set $component_values_cell_type $index [objarray get $component_values $element_vtk_id]
                                    incr index
                                }
                                lappend gidpost_values_cell_type $component_values_cell_type
                            }
                            set gidpost_ids_reorederd [objarray concat $gidpost_ids_reorederd $gidpost_ids_cell_type]
                            for { set i_component 0 } { $i_component < $num_components_used } { incr i_component } {
                                set prev_gipost_component_values [lindex $gidpost_values_reorederd $i_component]
                                set more_gipost_component_values [lindex $gidpost_values_cell_type $i_component]
                                lset gidpost_values_reorederd $i_component [objarray concat $prev_gipost_component_values $more_gipost_component_values]
                            }
                        }
                        set gidpost_gauss_point_name GP_ELEMENT_1
                        set gidpost_result [list Result $gidpost_resultname $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
                        lappend gidpost_result $gidpost_gauss_point_name
                        set gidpost_some_values [list $gidpost_ids_reorederd $gidpost_values_reorederd]
                        GiD_Result create -array $gidpost_result $gidpost_some_values
                        incr num_results
                    }
                } else {
                    set gidpost_result [list Result $gidpost_resultname $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
                    set gidpost_some_values [list $gidpost_ids $gidpost_values]
                    GiD_Result create -array $gidpost_result $gidpost_some_values
                    incr num_results
                }
            }
        }
        $vtkDataSetAttributes Delete
    }

    $vtkDataSet Delete

    return [list $num_results $num_graphs]
}

proc Vtk::ReadPostResults { filename analysis time_step } {
    if { ![file exists $filename] } {
        W [_ "File %s not found" $filename]
        return [list 0 0]
    }
    
    set is_vtk_dataset_reader 0
    set extension [string tolower [file extension $filename]]
    if { $extension == ".vtu" } {
        set h [vtkXMLUnstructuredGridReader New]
    } elseif { $extension == ".vtp" } {
        set h [vtkXMLPolyDataReader New]
    } elseif { $extension == ".vts" } {
        set h [vtkXMLStructuredGridReader New]
    } elseif { $extension == ".vtr" } {
        set h [vtkXMLRectilinearGridReader New]
    } elseif { $extension == ".vti" } {
        set h [vtkXMLImageDataReader New]
    } elseif { $extension == ".vto" } {
        set h [vtkXMLHyperOctreeReader New]
    } else {
        #.vtk
        set h [vtkDataSetReader New]
        set is_vtk_dataset_reader 1
    }
    $h SetFileName $filename
    $h GlobalWarningDisplayOff ;#to avoid error message with file 'financial.vtk' vtkStreamingDemandDrivenPipeline : Algorithm vtkDataSetReader returned failure for request: vtkInformation
    if { $is_vtk_dataset_reader } {
        foreach item {Scalars Vectors Normals Tensors TCoords ColorScalars Fields} {
            $h SetReadAll$item 1
        }
    }
    $h Update
    if { [catch { set res [Vtk::ReadPostResultArraySet $h $analysis $time_step] } msg] } {
        set res [list 0 0]
        W [string range $msg 0 1024]
    }
    $h Delete
    return $res
}

proc Vtk::GetFilesRecursive { dirname } {
    set names [glob -nocomplain -directory $dirname -types f *.vtk]
    foreach name [glob -nocomplain -directory $dirname -types d *] {
        lappend names {*}[Vtk::GetFilesRecursive $name]
    }
    return $names
}

# I assume that N is a decimal number
# ParaView recognizes file series named using certain patterns including:
# fooN.vtk
# foo_N.vtk
# foo-N.vtk
# foo.N.vtk
# Nfoo.vtk
# N.foo.vtk
# foo.vtk.N
# foo.vtk-sN
#try to find a pattern encoded in the name to know the time step

#kratos use other patterns more (problematics because can use float numbers and the dot seems a extension separator)
# foo_F.vtk
# F is a float number. Eexpected possitive and fixed point (not floating point) representation, in particular could be an integer N), and foo is arbitrary and cand contain also numbers
# e.g.
# model_part_0_0.60.vtk
# model_part_0_2.vtk

# how to know then the meaning of something like foo.3.vtk, it become not well defined!!
# with ParaView pattern foo.N.vtk -> N=3
# with kratos   pattern foo_F.vtk -> F=.3

# and extensions are not only vtk, but .vtk .vtu .vtp .vts .vtr .vti .vto

# this proc is only for internal test
proc Vtk::GetTimeStepFromName_Test { } {
    set filenames {kk_0_07.vtk kk_0_08.vtk foo3.vtk foo_3.vtk foo-3.vtk foo.3.vtk 3foo.vtk 3.foo.vtk foo.vtk.3 foo.vtk-s3 model_part_0_0.60.vtk model_part_0_2.vtk}
    foreach filename $filenames {
        W "$filename: [Vtk::GetTimeStepFromName $filename]"
    }
}

proc Vtk::GetTimeStepFromName { filename } {
    set extensions {.vtk .vtu .vtp .vts .vtr .vti .vto}
    set file_tail [file tail $filename]
    set extension [string tolower [file extension $file_tail]]
    set rootname [file rootname $file_tail]
    set base_name $rootname
    set time_step ""
    #warning [string is double 08] is false because 0 is octal prefix
    #then remove left zeros because before use 'string is double' or 'string is integer'
    #if future Tcl 9 octal numbers must begin with 0o and then this problem won't happen
    if { [lsearch $extensions $extension] != -1 } {
        # try kratos pattern
        # foo_F.vtk
        # also match some Paraview patterns
        # fooN.vtk
        # foo_N.vtk
        # foo-N.vtk
        # missunderstand this pattern foo.N.vtk , instead of decimal N gets float F=.N
        set n [string length $rootname]
        for { set i [expr $n-1] } { $i>0 } { incr i -1 } {
            set last_part [string trimleft [string range $rootname $i end] 0]
            #the '-' is not considered as the sign of a negative number, because time_step must be possitive, it is considered as a kind of separator
            if { ![string is double $last_part] || [string index $last_part 0] == "-" } {
                set base_name [string range $rootname 0 $i]
                set base_name_last [string index $base_name end]
                if { $base_name_last == "_" || $base_name_last == "-"} {
                    set base_name [string range $base_name 0 end-1]
                }
                set time_step [string range $rootname $i+1 end]
                break
            }
        }
        if  { $time_step == "" } {
            set extension2 [string tolower [file extension $rootname]]
            if { $extension2 != "" } {
                # N.foo.vtk
                if { [string is integer [string trimleft [file rootname $rootname] 0]] } {
                    set base_name [string range $extension2 1 end]
                    set time_step [file rootname $rootname]
                } else {
                    set time_step ""
                }
            } else {
                # Nfoo.vtk
                for { set i 0 } { $i<$n } { incr i } {
                    set first_part [string trimleft [string range $rootname 0 $i] 0]
                    if { ![string is integer $first_part] } {
                        set base_name [string range $rootname $i end]
                        set time_step [string range $rootname 0 $i-1]
                        break
                    }
                }
            }
        }
    } elseif { [lsearch $extensions [string range $extension 0 3]] != -1 && [string range $extension 4 5] == "-s" } {
        # foo.vtk-sN
        if { [string is integer [string trimleft [string range $extension 6 end]]] } {
            set base_name $rootname
            set time_step [string range $extension 6 end]
        } else {
            set time_step ""
        }
    } else {
        set extension2 [string tolower [file extension $rootname]]
        if { $extension2 != "" } {
            if { [lsearch $extensions $extension2] != -1 } {
                # foo.vtk.N
                if { [string is integer [string trimleft [string range $extension 1 end] 0]] } {
                    set base_name [file rootname $rootname]
                    set time_step [string range $extension 1 end]
                } else {
                    set time_step ""
                }
            } else {
                set time_step ""
            }
        }
    }

    if { [string index $time_step 0] == "0" && [string index $time_step 1] != "." } {
        #final value remove left zeros because [string is double 08] is false because 0 is octal prefix
        #but do not remove it in fixed point double case
        #e.g.
        # 008 -> 8
        # 0.8 -> 0.8
        set time_step [string trimleft $time_step 0]
    }
    return [list $base_name $time_step]
}

proc Vtk::ReadPost0 { filenames } {
    variable vtk_version
    variable _offsets_mesh_post ;#to associate results to the mesh ids
    variable _cell_multiple_num_equivalent_elements;#to associate results to mesh ids in case of elements that become multiple elements
    variable elements_vtk_id ;#to handle result on diferent element types that must be reordered by types in GiD
    variable mesh_bounds; #to create only once each mesh for different files with different time steps
    variable percent_done
    variable stop

    set vtk_version [package require vtk]
    package require objarray 1.2 ;#1.2 to use sort -unique
    package require vtk_objarray 1.1

    unset -nocomplain _offsets_mesh_post
    unset -nocomplain _cell_multiple_num_equivalent_elements
    unset -nocomplain elements_vtk_id
    unset -nocomplain mesh_bounds

    lassign {0 0 0} num_meshes num_results num_graphs
    
    set time_steps [list]
    set list_analysis [list]

    if { [llength $filenames] == 1 } {
        set filename [lindex $filenames 0]
        if { [file isdirectory $filename] } {
            #now the window only select files, not directories, but it is possible to invoke with folder
            #from Tcl command
            set filenames [Vtk::GetFilesRecursive $filename]
        } elseif { [string tolower [file extension $filename]] == ".pvd" } {
            set file_dirname [file dirname $filename]
            set filenames [list]
            package require tdom
            set xml [tDOM::xmlReadFile $filename]
            set doc [dom parse $xml]
            set dom_root [$doc documentElement]
            if { [$dom_root nodeName] != "VTKFile" } {
                W [_ "'%s' Is not a VTKFile file" $filename]
            } else {
                foreach dom_node [$dom_root selectNodes "Collection/DataSet"] {
                    set file [$dom_node getAttribute "file"]
                    if { [file pathtype $file] == "relative" } {
                        set file [file join $file_dirname $file]
                    }
                    lappend filenames $file
                    set timestep [$dom_node getAttribute "timestep"]
                    lappend time_steps $timestep
                    set group [$dom_node getAttribute "group"]
                    lappend list_analysis $group
                    #set part [$dom_node getAttribute "part"]
                }
            }
            $doc delete
        }
    }
    set ::Vtk::stop 0
    set tasks_max [expr {[llength $filenames]*2}]
    set tasks_done 0
    set ::Vtk::percent_done 0
    ::GidUtils::CreateAdvanceBar [_ "Reading file"] [_ "Process"] ::Vtk::percent_done 100 {set ::Vtk::stop 1}
    if { [llength $time_steps] != [llength $filenames] } {
        set time_steps [lrepeat [llength $filenames] ""]
    }
    if { [llength $list_analysis] != [llength $filenames] } {
        set list_analysis [lrepeat [llength $filenames] ""]
    }    
    foreach filename $filenames analysis $list_analysis  time_step $time_steps {
        if { $::Vtk::stop } {
            break
        }
        #current vtk version don't support non-ASCII filenames
        set tmp_filename ""
        if { ![string is ascii $filename] } {
            if { $::tcl_platform(platform) == "windows" } {
                set filename [file attributes $filename -shortname]
            } else {
                set tmp_filename [GidUtils::GetTmpFilename [file extension $filename]]
                file copy $filename $tmp_filename
                set filename $tmp_filename
            }
        }
        if { [file exists $filename] } {
            set err [ catch  {
                incr num_meshes [Vtk::ReadPostMeshes $filename]
            } err_txt]
            if { $err} {
                # error importing meshes, for instance when hitting academic mesh limits
                #on error must delete bar!!
                GidUtils::DeleteAdvanceBar

                unset -nocomplain _offsets_mesh_post
                unset -nocomplain mesh_bounds
                return -code error $err_txt
            }
            incr tasks_done
            set ::Vtk::percent_done [expr {int(($tasks_done*100.0)/$tasks_max)}]
            lassign [Vtk::ReadPostResults $filename $analysis $time_step] num_results_local num_graphs_local
            incr num_results $num_results_local
            incr num_graphs $num_graphs_local
            incr tasks_done
            set ::Vtk::percent_done [expr {int(($tasks_done*100.0)/$tasks_max)}]

            GiD_RaiseEvent GiD_Event_AfterOpenFile $filename VTK_FORMAT 0
            GiD_RaiseEvent GiD_Event_AfterChangeMesh [ GiD_Info Mesh NumNodes] [ GiD_Info Mesh NumElements]
            GiD_RaiseEvent GiD_Event_BeforeUpdateWindow SETS

        } else {
            incr tasks_done 2
            set ::Vtk::percent_done [expr {int(($tasks_done*100.0)/$tasks_max)}]
        }
        if { $tmp_filename != "" } {
            file delete $tmp_filename
        }

    }

    #on error must delete bar!!
    GidUtils::DeleteAdvanceBar

    unset -nocomplain _offsets_mesh_post
    unset -nocomplain mesh_bounds

    return [list $num_meshes $num_results $num_graphs]
}

#read and finally redraw showing what is read
proc Vtk::ReadPost { filenames { lst_options {}}} {
    GidUtils::WaitState .gid
    GidUtils::DisableGraphics

    # process lst_options
    if { [ llength $lst_options] > 0} {
        array set my_options $lst_options
        if { [ info exists my_options(merge)]} {
            set Vtk::_merge $my_options(merge)
        }
    }

    if { !$Vtk::_merge } {
        GiD_Process Mescape Files New Yes escape ;#to avoid previous geometry
    }

    if { [catch { set res [Vtk::ReadPost0 $filenames] } msg] } {
        set num_meshes 0
        set num_results 0
        set num_graphs 0
        W [string range $msg 0 1024]
    } else {
        lassign $res num_meshes num_results num_graphs
    }
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    if { $num_meshes || $num_results > 0 } {
        GiD_Process Mescape View Zoom Frame escape
    }
    if { $num_results > 0 } {
        #set the first analysis and step to use
        set analysis_to_use [lindex [GiD_Info postprocess get all_analysis] 0]
        set step_to_use [lindex [GiD_Info postprocess get all_steps $analysis_to_use] 0]
        GiD_Process Mescape results analysissel $analysis_to_use $step_to_use escape
    } elseif { $num_graphs > 0 } {
        GiD_Graph show
    }
    GidUtils::SetWarnLine [_ "Read %d meshes and %d results %d graphs" $num_meshes $num_results $num_graphs]
    return
}

#to drag and drop a file
proc Vtk::ReadPostSingleFile { filename } {
    variable _merge
    GidUtils::EvalAndEnterInBatch Vtk::ReadPost [list $filename] [list merge $Vtk::_merge]
}

#to add more options to the standar file browser
proc Vtk::MoreImportOptions { f } {
    variable _merge
    ttk::checkbutton $f.chkGeometry -text [_ "Merge"] -variable Vtk::_merge
    grid columnconfigure $f 1 -weight 1
    grid $f.chkGeometry -sticky w -columnspan 2

    variable export_dir
    if { ![info exists export_dir] } {
        #trick to show the more options
        set export_dir open
    }
    return Vtk::export_dir
}

proc Vtk::ReadPostW { } {
    variable _merge
    set multiple 1
    set filenames [MessageBoxGetFilename file read [_ "Read Vtk results file"] \
        {} {{{Vtk} {.vtk .vtu .vtp .vts .vtr .vti .pvd}} {{All files} {.*}}} {} $multiple \
        [list [_ "Import options"] Vtk::MoreImportOptions]]
    #.vti .vto
    if {$filenames == ""} {
        return
    }
    GidUtils::EvalAndEnterInBatch Vtk::ReadPost $filenames [ list merge $Vtk::_merge]
}

#########################################################################################
######################################## WRITE POST #####################################
#########################################################################################

proc Vtk::ExistsVolumicMesh { } {
    set exists_volumic_mesh 0
    foreach element_type {Tetrahedra Hexahedra Prism Pyramid Sphere} {
        if { [GiD_Info Mesh -post NumElements $element_type] } {
            set exists_volumic_mesh 1
            break
        }
    }
    return $exists_volumic_mesh
}

proc Vtk::WritePostMeshAndResults { filename } {
    variable vtk_version
    variable percent_done
    variable _data_mode
    variable _all_result_steps
    variable _all_results

    set fail 0

    if { ![GiD_Info Mesh -post NumNodes] } {
        return 1
    }

    #calculate num_files to write based of results (each time step must be in a different file)
    set num_files 0
    if { $_all_results } {
        set analysis_names [GiD_Info postprocess get all_analysis]
    } else {
        set analysis_names [list [GiD_Info postprocess get cur_analysis]]
    }
    foreach analysis_name $analysis_names {
        if { $_all_result_steps } {
            set time_steps [GiD_Info postprocess get all_steps $analysis_name]
        } else {
            set time_steps [list [GiD_Info postprocess get cur_step]]
        }
        incr num_files [llength $time_steps]
    }

    set exists_volumic_mesh [Vtk::ExistsVolumicMesh]
    if { $exists_volumic_mesh  } {
        set h [vtkXMLUnstructuredGridWriter New]
        #set .vtu extension
        set filename [file rootname $filename].vtu
    } else {
        #if mesh has only lines, triangles and quads can use PolyData
        set h [vtkXMLPolyDataWriter New]
        #set .vtp extension
        set filename [file rootname $filename].vtp
    }

    $h SetFileName $filename
    if { $_data_mode == "Ascii" } {
        $h SetDataModeToAscii
    } elseif { $_data_mode == "Binary" } {
        $h SetDataModeToBinary
        #$h SetCompressorTypeToNone
        $h SetCompressorTypeToZLib ;#to compress Binary and Appended Data
    } elseif { $_data_mode == "Compressed" } {
        $h SetDataModeToAppended
        #$h SetCompressorTypeToNone
        $h SetCompressorTypeToZLib ;#to compress Binary and Appended Data
        #AppendedData encoding="base64" en los ejemplos KwiViewer es "raw"
        #$h EncodeAppendedDataOn ;#AppendedData encoding="base64". XML compliant but slower
        $h EncodeAppendedDataOff ;#AppendedData encoding="raw". NON XML compliant but faster
    } else {
        W "Unexpected data mode $_data_mode"
    }

    if { $num_files == 0 } {
        #doesn't has any result, write only the mesh that is not associated to time steps of results, could use -step 0 or avoid this flag
        if { $exists_volumic_mesh  } {
            set vtkUnstructuredGrid [vtkUnstructuredGrid New]
            Vtk::PostMeshFillvtkUnstructuredGrid 0 $vtkUnstructuredGrid
            set vtkDataSet $vtkUnstructuredGrid
        } else {
            set vtkPolyData [vtkPolyData New]
            Vtk::PostMeshFillvtkPolyData 0 $vtkPolyData
            set vtkDataSet $vtkPolyData
        }
        if { $vtk_version == 5.8 } {
            $h SetInput $vtkDataSet
        } else {
            #changed in vtk 6.1
            $h SetInputData $vtkDataSet
        }

        $h Write
        $h Delete
        GiD_RaiseEvent GiD_Event_AfterSaveFile $filename VTK_FORMAT 0
        return $fail
    }

    set i_file 0

    set ::Vtk::stop 0
    set ::Vtk::percent_done 0
    ::GidUtils::CreateAdvanceBar [_ "Writting file"] [_ "Process"] ::Vtk::percent_done 100 {set ::Vtk::stop 1}


    set mesh_variable_along_time 1
    if { [GidUtils::VersionCmp 12.1.5d] >= 0 } {
        set mesh_variable_along_time [GiD_Info postprocess get is_mesh_variable_along_steps]
    }

    set doc_pvd ""
    set collection_node ""
    if { $num_files>1 } {
        set filename_pvd [file rootname $filename].pvd
        set doc_pvd [dom parse "<VTKFile type='Collection' version='0.1'></VTKFile>"]
        set root_pvd [$doc_pvd documentElement]
        set collection_node [Vtk::XmlCreateElement $root_pvd "Collection" ""]
    }
    #a vtk file can't store several time steps (without tricks like encode in the dataset name)
    set num_results 0
    if { $_all_results } {
        set analysis_names [GiD_Info postprocess get all_analysis]
    } else {
        set analysis_names [list [GiD_Info postprocess get cur_analysis]]
    }
    foreach analysis_name $analysis_names {
        if { $_all_result_steps } {
            set time_steps [GiD_Info postprocess get all_steps $analysis_name]
        } else {
            set time_steps [list [GiD_Info postprocess get cur_step]]
        }
        set i_step 0
        foreach time_step $time_steps {
            if { $::Vtk::stop } {
                break
            }
            if { $mesh_variable_along_time || ![ info exists vtkDataSet]} {
                if { $exists_volumic_mesh  } {
                    set vtkUnstructuredGrid [vtkUnstructuredGrid New]
                    Vtk::PostMeshFillvtkUnstructuredGrid $i_step $vtkUnstructuredGrid
                    set vtkDataSet $vtkUnstructuredGrid
                } else {
                    set vtkPolyData [vtkPolyData New]
                    Vtk::PostMeshFillvtkPolyData $i_step $vtkPolyData
                    set vtkDataSet $vtkPolyData
                }
                $vtkDataSet GlobalReleaseDataFlagOff ;#to not clear it in next time steps (maybe has been set to On reading vtk)
                if { $vtk_version == 5.8 } {
                    $h SetInput $vtkDataSet
                } else {
                    #changed in vtk 6.1
                    $h SetInputData $vtkDataSet
                }
            } else {
                #mesh is constant for all files, then must clear $vtkDataSet of previous results
                set vtkPointData [$vtkDataSet GetPointData]
                $vtkPointData Initialize
                set vtkCellData [$vtkDataSet GetCellData]
                $vtkCellData Initialize
            }
            if { $_all_results } {
                set result_names [GiD_Info postprocess get results_list CONTOUR_FILL $analysis_name $time_step]
            } else {
                set result_names [list [GiD_Info postprocess get cur_result]]
                if { $result_names == "" } {
                    W "Must set a current result to be exported"
                }
            }
            foreach result_name $result_names {
                Vtk::PostResultFillvtkDataSet $analysis_name $result_name $time_step $vtkDataSet
                incr num_results
            }
            if { $num_files>1 } {
                set filename_i [file rootname $filename][format "%04d" $i_file][file extension $filename]
                $h SetFileName $filename_i
                if { $collection_node != "" } {
                    Vtk::XmlCreateElement $collection_node "DataSet" [list timestep $time_step group $analysis_name part 0 file [file tail $filename_i]]
                }
            } else {
                set filename_i $filename
            }
            $h Write
            GiD_RaiseEvent GiD_Event_AfterSaveFile $filename_i VTK_FORMAT 0
            incr i_file
            set ::Vtk::percent_done [expr {int(($i_file*100.0)/$num_files)}]
            incr i_step
        }
    }
    
    if { $doc_pvd != "" } {
        set filename_pvd [file rootname $filename].pvd
        Vtk::SaveXmlFile $doc_pvd $filename_pvd
    }

    if { $num_files == 0 } {
        $h Write
        GiD_RaiseEvent GiD_Event_AfterSaveFile $filename VTK_FORMAT 0
        incr num_file
        incr ::Vtk::percent_done
    }
    GidUtils::DeleteAdvanceBar

    $h Delete
    return $fail
}

proc Vtk::PostMeshFillvtkPolyData { i_step vtkPolyData } {
    variable VtkElementType
    variable GidPostElementNNode

    set nodes [GiD_Info mesh -post -step_index $i_step nodes -array]
    lassign $nodes ids coordinates
    set num_nodes [objarray length $ids]
    lassign $coordinates xs ys zs

    #node numbers are implicit from 0 and consecutives in vtk format
    set vtkPoints [vtkPoints New]
    $vtkPoints SetDataTypeToFloat
    set objarray [objarray concat $xs $ys $zs]
    objarray interleave $objarray 3
    set old_ids $ids
    set vtkDataArray [vtk_objarray vtk_data_array_from_objarray $objarray vtkFloatArray]
    $vtkDataArray SetNumberOfComponents 3
    $vtkPoints SetData $vtkDataArray
    $vtkPolyData SetPoints $vtkPoints

    # getting 'Point' elements
    set vtkCellArray [vtkCellArray New]
    set objarray [objarray new intarray 0]
    set num_elements 0
    set element_type Point
    set vtk_element_type $VtkElementType($element_type)
    set element_nnodes $GidPostElementNNode($vtk_element_type)
    set elements [GiD_Info mesh -post -step_index $i_step elements $element_type -array]
    #could return the mesh of same kind of element as a list with more of multiple parts!!
    foreach sub_mesh $elements {
        lassign $sub_mesh elemtype ids connectivities materials
        set num_elements_local [objarray length $ids]
        if { $num_elements_local } {
            incr num_elements $num_elements_local
            #set element_nnodes [llength $connectivities] ;#wrong for quadratic case...
            set connectivities [lrange $connectivities 0 $element_nnodes-1] ;#write quadratic as linear
            set array_nnodes [objarray new intarray $num_elements_local $element_nnodes]
            set array_connectivities [objarray concat {*}$connectivities]
            objarray renumber $array_connectivities $old_ids
            set objarray_local [objarray concat $array_nnodes $array_connectivities]
            objarray interleave $objarray_local [expr {$element_nnodes+1}]
            set objarray [objarray concat $objarray $objarray_local]
        }
    }
    if { [ llength $elements] != 0} {
        #fill objarray
        set vtkIdTypeArray [vtk_objarray vtk_data_array_from_objarray $objarray vtkIdTypeArray]
        $vtkCellArray SetCells $num_elements $vtkIdTypeArray
        $vtkPolyData SetVerts $vtkCellArray
    }

    # getting 'Line' elements
    set vtkCellArray [vtkCellArray New]
    set objarray [objarray new intarray 0]
    set num_elements 0
    set element_type Line
    set vtk_element_type $VtkElementType($element_type)
    set element_nnodes $GidPostElementNNode($vtk_element_type)
    set elements [GiD_Info mesh -post -step_index $i_step elements $element_type -array]
    #could return the mesh of same kind of element as a list with more of multiple parts!!
    foreach sub_mesh $elements {
        lassign $sub_mesh elemtype ids connectivities materials
        set num_elements_local [objarray length $ids]
        if { $num_elements_local } {
            incr num_elements $num_elements_local
            #set element_nnodes [llength $connectivities] ;#wrong for quadratic case...
            set connectivities [lrange $connectivities 0 $element_nnodes-1] ;#write quadratic as linear
            set array_nnodes [objarray new intarray $num_elements_local $element_nnodes]
            set array_connectivities [objarray concat {*}$connectivities]
            objarray renumber $array_connectivities $old_ids
            set objarray_local [objarray concat $array_nnodes $array_connectivities]
            objarray interleave $objarray_local [expr {$element_nnodes+1}]
            set objarray [objarray concat $objarray $objarray_local]
        }
    }
    if { [ llength $elements] != 0} {
        #fill objarray
        set vtkIdTypeArray [vtk_objarray vtk_data_array_from_objarray $objarray vtkIdTypeArray]
        $vtkCellArray SetCells $num_elements $vtkIdTypeArray
        $vtkPolyData SetLines $vtkCellArray
    }

    # getting 'Triangle' and 'Quadrilateral' elements
    #must add Triangle and Quadrilateral to the same vtkCellArray
    set vtkCellArray [vtkCellArray New]
    set objarray [objarray new intarray 0]
    set num_elements 0
    foreach element_type { Triangle Quadrilateral } {
        set vtk_element_type $VtkElementType($element_type)
        set element_nnodes $GidPostElementNNode($vtk_element_type)
        set elements [GiD_Info mesh -post -step_index $i_step elements $element_type -array]
        #could return the mesh of same kind of element as a list with more of multiple parts!!
        foreach sub_mesh $elements {
            lassign $sub_mesh elemtype ids connectivities materials
            set num_elements_local [objarray length $ids]
            incr num_elements $num_elements_local
            #set element_nnodes [llength $connectivities] ;#wrong for quadratic case...
            set connectivities [lrange $connectivities 0 $element_nnodes-1] ;#write quadratic as linear
            set array_nnodes [objarray new intarray $num_elements_local $element_nnodes]
            set array_connectivities [objarray concat {*}$connectivities]
            objarray renumber $array_connectivities $old_ids
            set objarray_local [objarray concat $array_nnodes $array_connectivities]
            objarray interleave $objarray_local [expr {$element_nnodes+1}]
            set objarray [objarray concat $objarray $objarray_local]
        }
    }
    if { $num_elements != 0} {
        #fill objarray
        set vtkIdTypeArray [vtk_objarray vtk_data_array_from_objarray $objarray vtkIdTypeArray]
        $vtkCellArray SetCells $num_elements $vtkIdTypeArray
        $vtkPolyData SetPolys $vtkCellArray
        #$vtkCellArray Delete
    }
    return 0
}

proc Vtk::PostMeshFillvtkUnstructuredGrid { i_step vtkUnstructuredGrid } {
    variable VtkElementType
    variable GidPostElementNNode

    set nodes [GiD_Info mesh -post -step_index $i_step nodes -array]
    lassign $nodes ids coordinates
    set num_nodes [objarray length $ids]
    lassign $coordinates xs ys zs

    #node numbers are implicit from 0 and consecutives in vtk format
    set vtkPoints [vtkPoints New]
    $vtkPoints SetDataTypeToFloat
    set objarray [objarray concat $xs $ys $zs]
    objarray interleave $objarray 3
    set old_ids $ids
    set vtkDataArray [vtk_objarray vtk_data_array_from_objarray $objarray vtkFloatArray]
    $vtkDataArray SetNumberOfComponents 3
    $vtkPoints SetData $vtkDataArray
    $vtkUnstructuredGrid SetPoints $vtkPoints


    #must add all elements to the same vtkCellArray
    set vtkCellArray [vtkCellArray New]
    set vtkUnsignedCharArray [vtkUnsignedCharArray New]
    set objarray [objarray new intarray 0]
    set num_elements 0
    foreach element_type { Point Line Triangle Quadrilateral Circle Tetrahedra Hexahedra Prism Pyramid Sphere } {
        set vtk_element_type $VtkElementType($element_type)
        set element_nnodes $GidPostElementNNode($vtk_element_type)
        set elements [GiD_Info mesh -post -step_index $i_step elements $element_type -array]
        #could return the mesh of same kind of element as a list with more of multiple parts!!
        foreach sub_mesh $elements {
            lassign $sub_mesh elemtype ids connectivities materials
            set num_elements_local [objarray length $ids]
            incr num_elements $num_elements_local
            #set element_nnodes [llength $connectivities] ;#wrong for quadratic case...
            set connectivities [lrange $connectivities 0 $element_nnodes-1] ;#to write quadratic as linear
            set array_nnodes [objarray new intarray $num_elements_local $element_nnodes]
            set array_connectivities [objarray concat {*}$connectivities]
            objarray renumber $array_connectivities $old_ids
            set objarray_local [objarray concat $array_nnodes $array_connectivities]
            objarray interleave $objarray_local [expr {$element_nnodes+1}]
            set objarray [objarray concat $objarray $objarray_local]
            for {set i_element 0} {$i_element < $num_elements_local} {incr i_element} {
                $vtkUnsignedCharArray InsertNextValue $vtk_element_type
            }
        }
    }
    #fill objarray
    set vtkIdTypeArray [vtk_objarray vtk_data_array_from_objarray $objarray vtkIdTypeArray]
    $vtkCellArray SetCells $num_elements $vtkIdTypeArray
    #$vtkIdTypeArray Delete

    #element type can change
    #$vtkUnstructuredGrid SetCells [$vtkUnsignedCharArray GetValue 0] $vtkCellArray
    set vtkIdTypeArrayDummy [vtkIdTypeArray New]
    $vtkUnstructuredGrid SetCells $vtkUnsignedCharArray $vtkIdTypeArrayDummy $vtkCellArray
    #vtkIdTypeArrayDummy Delete
    #$vtkCellArray Delete
    return 0
}
proc Vtk::PostResultFillvtkDataSet { analysis_name result_name time_step vtkDataSet } {
    set num_results 0

    set gidpost_type ""
    set gidpost_over ""
    set gidpost_unit ""
    set gidpost_component_names [list]
    set result_info [GiD_Result get -array [list $result_name $analysis_name $time_step]]
    foreach item $result_info {
        set first_subitem [lindex $item 0]
        if { [objarray type $first_subitem] != "" } {
            #check objarray type for efficienct, to avoid objarray conversion to string
            #set result_data $item
            continue
        }
        if { $first_subitem == "Result"} {
            set gidpost_type [lindex $item 4]
            set gidpost_over [lindex $item 5]
        } elseif { $first_subitem == "Unit"} {
            set gidpost_unit [lindex $item 1]
        } elseif { $first_subitem == "ComponentNames"} {
            set gidpost_component_names [lrange $item 1 end]
        } else {
            #set result_data $item
        }
    }
    set result_data [lindex $result_info end]
    set gidpost_ids [lindex $result_data 0]
    if { $gidpost_over == "OnNodes" } {
        set vtkPointData [$vtkDataSet GetPointData]
        set vtkDataSetAttributes $vtkPointData
    } else {
        set vtkCellData [$vtkDataSet GetCellData]
        set vtkDataSetAttributes $vtkCellData
    }
    if { [llength $gidpost_component_names] } {
        set component_names $gidpost_component_names
    } else {
        if { $gidpost_type == "Scalar" } {
            set component_names [list x]
        } elseif { $gidpost_type == "Vector" } {
            set component_names [list x y z]
        } elseif { $gidpost_type == "Matrix" } {
            set component_names [list xx yy zz xy yz xz]
        } else {
            set component_names [list]
        }
    }
    set num_values [objarray length $gidpost_ids]

    set i_component 0
    foreach component_values [lindex $result_data 1] {
        set num_values_i [objarray length $component_values]
        if { $num_values_i != $num_values } {
            if { $gidpost_over == "OnNodes" } {
                W "Error result [list $result_name $analysis_name $time_step]. Result is defined for $num_values entities but are returned $num_values_i values. Used only firsts"
                set component_values [objarray range $component_values 0 [expr {$num_values-1}]]
            } else {
                #multiple gauss points, convert to one averaging values
                set num_gauss_points [expr {int($num_values_i)/int($num_values) }]
                set type [objarray type $component_values]
                set averaged_component_values [objarray new $type $num_values]
                set pos_ini 0
                set pos_end [expr {$num_gauss_points-1}]
                for {set i 0} {$i<$num_values} {incr i} {
                    set gauss_values [objarray range $component_values $pos_ini $pos_end]
                    set sum [objarray sum $gauss_values]
                    set mean [expr {double($sum)/double($num_gauss_points)}]
                    objarray set $averaged_component_values $i $mean
                    incr pos_ini $num_gauss_points
                    incr pos_end $num_gauss_points
                }
                set component_values $averaged_component_values
            }
        }
        set data($i_component) $component_values
        incr i_component
    }

    if { $gidpost_type == "Scalar" } {
        set vtkDataArray [vtk_objarray vtk_data_array_from_objarray $data(0) vtkFloatArray]
        #$vtkDataSetAttributes SetScalars $vtkDataArray
        $vtkDataSetAttributes AddArray $vtkDataArray
    } else {
        set num_components [llength [array names data]]
        if { $gidpost_type == "Vector" } {
            if { $num_components == 4 } {
                #neglect the 4th component of vector (modulus) that is redundant
                set num_components 3
                set component_names [lrange $component_names 0 2]
            }
        }
        if { $gidpost_type == "Vector" } {
            set data_joined [objarray concat $data(0) $data(1) $data(2)]
            objarray interleave $data_joined 3
        } else {
            set data_list [list]
            for {set i_component 0} {$i_component<$num_components} {incr i_component} {
                lappend data_list $data($i_component)
            }
            set data_joined [objarray concat {*}$data_list]
            objarray interleave $data_joined $num_components
        }
        set vtkDataArray [vtk_objarray vtk_data_array_from_objarray $data_joined vtkFloatArray]
        $vtkDataArray SetNumberOfComponents $num_components

        if { [llength $component_names] != $num_components } {
            if { $gidpost_type == "Matrix" && [llength $component_names] == [expr {$num_components-3}]} {
                #seems that eigenvalues are provided but not their names
                lappend component_names c-I c-II c-III
            }
        }
        if { [llength $component_names] != $num_components } {
            W "WritePostResult: component_names='$component_names' but num_components=$num_components. Use numbers for component names"
            for {set i_component 0} {$i_component<$num_components} {incr i_component} {
                $vtkDataArray SetComponentName $i_component "c-[expr {$i_component+1}]"
            }
        } else {
            for {set i_component 0} {$i_component<$num_components} {incr i_component} {
                $vtkDataArray SetComponentName $i_component [lindex $component_names $i_component]
            }
        }
        #$vtkDataSetAttributes SetVectors $vtkDataArray
        $vtkDataSetAttributes AddArray $vtkDataArray
    }

    $vtkDataArray SetName $result_name
    #$vtkDataArray SetName [list $analysis_name $result_name $time_step]
    incr num_results
    return $num_results
}

proc Vtk::WritePost0 { filename } {
    variable vtk_version
    set vtk_version [package require vtk]
    package require objarray 1.7 ;#1.7 to use objarray renumber
    package require vtk_objarray 1.1

    set fail [Vtk::WritePostMeshAndResults $filename]
    return $fail
}

proc Vtk::WritePost { filename } {
    variable percent_done
    GidUtils::WaitState .gid
    GidUtils::DisableGraphics
    set fail 0
    set err [catch { set fail [Vtk::WritePost0 $filename] } msg]
    if { $err } {
        set Vtk::percent_done 100.0
        set fail 1
        W $msg
    }
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    if { $fail } {
        GidUtils::SetWarnLine [_ "Problems writting"]
    } else {
        GidUtils::SetWarnLine [_ "File '%s' written" $filename]
    }
    return $fail
}

#to add more options to the standar file browser
proc Vtk::MoreExportOptions { f } {
    variable _all_results
    variable _all_result_steps
    variable _data_mode
    ttk::label $f.lDataMode -text [_ "Data mode"]
    set values {Ascii Binary Compressed}
    set labels [list [_ "ASCII"] [_ "Binary"] [_ "Compressed"]]
    TTKComboBox $f.cbDataMode -textvariable Vtk::_data_mode -labels $labels -values $values -state readonly
    ttk::checkbutton $f.chkAllResults -text [_ "All results"] -variable Vtk::_all_results
    ttk::checkbutton $f.chkAllResultSteps -text [_ "All steps"] -variable Vtk::_all_result_steps
    grid columnconfigure $f 1 -weight 1
    grid $f.lDataMode $f.cbDataMode -sticky w
    grid $f.chkAllResults $f.chkAllResultSteps -sticky w ;#-columnspan 2

    variable export_dir
    if { ![info exists export_dir] } {
        #trick to show the more options
        set export_dir open
    }
    return Vtk::export_dir
}

proc Vtk::WritePostW { } {
    set filename [MessageBoxGetFilename file write [_ "Write Vtk results file"] \
        {} {{{Vtk} {.vtp .vtu}} {{All files} {.*}}} {} 0\
        [list [_ "Export options"] Vtk::MoreExportOptions]]
    #.vti .vto
    if {$filename == ""} {
        return
    }
    GidUtils::EvalAndEnterInBatch Vtk::WritePost $filename
}

#########################################################################################
######################################### OTHER #########################################
#########################################################################################

proc Vtk::AddToMenuPOST { } {
    if { [GidUtils::IsTkDisabled] } {
        return
    }
    if { [GiDMenu::GetOptionIndex Files [list Import Vtk...] POST] != -1 } {
        return
    }
    #try to insert this menu after the word "Files->Import->XYZ nodes"
    set position [GiDMenu::GetOptionIndex Files [list Import "XYZ nodes"] POST]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Files [list Import Vtk...] $position POST Vtk::ReadPostW "" "" insertafter _

    #try to insert this menu after the word "Files->Export->Visible surfaces to PLY"
    set position [GiDMenu::GetOptionIndex Files [list Export "Visible surfaces to PLY"] POST]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Files [list Export Vtk...] $position POST Vtk::WritePostW "" "" insertafter _
}

proc Vtk::AddToMenu { } {
    Vtk::AddToMenuPOST
}

Vtk::Init

#invoke this menus changes
Vtk::AddToMenu
GiDMenu::UpdateMenus

#register the proc to be automatically called when re-creating all menus (e.g. when doing files new)
GiD_RegisterPluginAddedMenuProc Vtk::AddToMenu

#register the proc to be automatically called when dropping a file
foreach extension {.vtk .vtu .vtp .vts .vtr .vti .pvd} {
    GiD_RegisterExtensionProc $extension POST Vtk::ReadPostSingleFile
}

#Vtk::ReadPost0 {C:\gid project more\Graficos\VTK\Open FOAM example\VTK}
